"""Permission service for SkyPilot API Server."""
import contextlib
import hashlib
import logging
import os
import threading
from typing import Generator, List, Optional

import casbin
import filelock
import sqlalchemy_adapter

from sky import global_user_state
from sky import models
from sky import sky_logging
from sky.skylet import constants
from sky.users import rbac
from sky.utils import annotations
from sky.utils import common
from sky.utils import common_utils
from sky.utils.db import db_utils

logging.getLogger('casbin.policy').setLevel(sky_logging.ERROR)
logging.getLogger('casbin.role').setLevel(sky_logging.ERROR)
logging.getLogger('casbin.model').setLevel(sky_logging.ERROR)
logging.getLogger('casbin.rbac').setLevel(sky_logging.ERROR)
logger = sky_logging.init_logger(__name__)

# Filelocks for the policy update.
POLICY_UPDATE_LOCK_PATH = os.path.expanduser('~/.sky/.policy_update.lock')
POLICY_UPDATE_LOCK_TIMEOUT_SECONDS = 20

_enforcer_instance: Optional['PermissionService'] = None


class PermissionService:
    """Permission service for SkyPilot API Server."""

    def __init__(self):
        self.enforcer: Optional[casbin.Enforcer] = None
        self._lock = threading.Lock()

    def initialize(self):
        self._lazy_initialize(full_initialize=True)

    def _lazy_initialize(self, full_initialize: bool = False):
        if self.enforcer is not None:
            return
        with self._lock:
            if self.enforcer is not None:
                return
            global _enforcer_instance
            if _enforcer_instance is None:
                engine = global_user_state.initialize_and_get_db()
                if full_initialize:
                    db_utils.add_all_tables_to_db_sqlalchemy(
                        sqlalchemy_adapter.Base.metadata, engine)
                adapter = sqlalchemy_adapter.Adapter(
                    engine, db_class=sqlalchemy_adapter.CasbinRule)
                model_path = os.path.join(os.path.dirname(__file__),
                                          'model.conf')
                enforcer = casbin.Enforcer(model_path, adapter)
                self.enforcer = enforcer
                # Only set the enforcer instance once the enforcer
                # is successfully initialized, if we change it and then fail
                # we will set it to None and all subsequent calls will fail.
                _enforcer_instance = self
                if full_initialize:
                    with _policy_lock():
                        self._maybe_initialize_policies()
                        self._maybe_initialize_basic_auth_user()
            else:
                assert _enforcer_instance is not None
                self.enforcer = _enforcer_instance.enforcer

    def _ensure_enforcer(self) -> casbin.Enforcer:
        """Ensure enforcer is initialized and return it."""
        self._lazy_initialize()
        assert self.enforcer is not None, (
            'Enforcer should be initialized after _lazy_initialize()')
        return self.enforcer

    def _get_plugin_rbac_rules(self):
        """Get RBAC rules from loaded plugins.

        Returns:
            Dictionary of plugin RBAC rules, or empty dict if plugins module
            is not available or no rules are defined.
        """
        try:
            # pylint: disable=import-outside-toplevel
            from sky.server import plugins as server_plugins
            return server_plugins.get_plugin_rbac_rules()
        except ImportError:
            # Plugin module not available (e.g., not running as server)
            logger.debug(
                'Plugin module not available, skipping plugin RBAC rules')
            return {}
        except Exception as e:  # pylint: disable=broad-except
            logger.warning(f'Failed to get plugin RBAC rules: {e}')
            return {}

    def _maybe_initialize_basic_auth_user(self) -> None:
        """Initialize basic auth user if it is enabled."""
        basic_auth = os.environ.get(constants.SKYPILOT_INITIAL_BASIC_AUTH)
        if not basic_auth:
            return
        username, password = basic_auth.split(':', 1)
        if username and password:
            user_hash = hashlib.md5(
                username.encode()).hexdigest()[:common_utils.USER_HASH_LENGTH]
            user_info = global_user_state.get_user(user_hash)
            if user_info:
                logger.debug(f'Basic auth user {username} already exists')
                return
            global_user_state.add_or_update_user(
                models.User(id=user_hash,
                            name=username,
                            password=password,
                            user_type=models.UserType.BASIC.value))
            enforcer = self._ensure_enforcer()
            enforcer.add_grouping_policy(user_hash, rbac.RoleName.ADMIN.value)
            enforcer.save_policy()
            logger.info(f'Basic auth user {username} initialized')

    def _maybe_initialize_policies(self) -> None:
        """Initialize policies if they don't already exist."""
        logger.debug(f'Initializing policies in process: {os.getpid()}')

        policy_updated = False

        # Check if policies are already initialized by looking for existing
        # permission policies in the enforcer
        enforcer = self._ensure_enforcer()
        # Convert existing policies to set of tuples for O(1) lookups
        existing_policies = {tuple(p) for p in enforcer.get_policy()}

        # Get plugin RBAC rules dynamically
        plugin_rules = self._get_plugin_rbac_rules()

        # If we already have policies for the expected roles, skip
        # initialization
        role_permissions = rbac.get_role_permissions(plugin_rules=plugin_rules)
        expected_policies = []
        for role, permissions in role_permissions.items():
            if permissions.get('permissions'
                              ) and 'blocklist' in permissions['permissions']:
                blocklist = permissions['permissions']['blocklist']
                for item in blocklist:
                    expected_policies.append(
                        (role, item['path'], item['method']))

        # Add workspace policy
        workspace_policy_permissions = rbac.get_workspace_policy_permissions()
        logger.debug(f'Workspace policy permissions from config: '
                     f'{workspace_policy_permissions}')

        for workspace_name, users in workspace_policy_permissions.items():
            for user in users:
                expected_policies.append((user, workspace_name, '*'))
        # Check if all expected policies already exist and find missing ones
        missing_policies = [
            p for p in expected_policies if p not in existing_policies
        ]
        # Find policies to remove
        expected_policies_set = set(expected_policies)
        redundant_policies = [
            p for p in existing_policies if p not in expected_policies_set
        ]
        if missing_policies:
            # Add missing policies
            logger.debug(f'Found {len(missing_policies)} missing policies, '
                         'initializing...')
            for p in missing_policies:
                logger.debug(f'Adding policy: {p}')
                enforcer.add_policy(*p)
                policy_updated = True
            logger.debug('Missing policies added successfully')

        if redundant_policies:
            # Remove redundant policies
            logger.debug(f'Found {len(redundant_policies)} redundant policies, '
                         'cleaning up...')
            for p in redundant_policies:
                logger.debug(f'Removing policy: {p}')
                enforcer.remove_policy(*p)
                policy_updated = True
            logger.debug('Redundant policies removed successfully')

        if not missing_policies and not redundant_policies:
            logger.debug('Policies already in sync, skipping initialization')

        # Always ensure users have default roles (this is idempotent)
        # Get users who already have roles (g policies) to avoid redundant calls
        users_with_roles = {tuple(g)[0] for g in enforcer.get_grouping_policy()}
        all_users = global_user_state.get_all_users()
        for existing_user in all_users:
            if str(existing_user.id) not in users_with_roles:
                logger.debug(f'Adding role for user: {existing_user.name}'
                             f'({existing_user.id})')
                user_added = self._add_user_if_not_exists_no_lock(
                    existing_user.id)
                policy_updated = policy_updated or user_added
        for system_user_id in [
                common.SERVER_ID, constants.SKYPILOT_SYSTEM_USER_ID
        ]:
            if system_user_id not in users_with_roles:
                logger.debug(f'Adding role for system user: {system_user_id}')
                user_added = self._add_user_if_not_exists_no_lock(
                    system_user_id, rbac.RoleName.ADMIN.value)
                policy_updated = policy_updated or user_added
        if policy_updated:
            enforcer.save_policy()

    def add_user_if_not_exists(self, user_id: str) -> None:
        """Add user role relationship."""
        self._lazy_initialize()
        with _policy_lock():
            self._add_user_if_not_exists_no_lock(user_id)

    def _add_user_if_not_exists_no_lock(self,
                                        user_id: str,
                                        role: Optional[str] = None) -> bool:
        """Add user role relationship without lock.

        Returns:
            True if the user was added, False otherwise.
        """
        enforcer = self._ensure_enforcer()
        user_roles = enforcer.get_roles_for_user(user_id)
        if not user_roles:
            enforcer.add_grouping_policy(user_id, role or
                                         rbac.get_default_role())
            return True
        return False

    def delete_user(self, user_id: str) -> None:
        """Delete user role relationship."""
        with _policy_lock():
            # Get current roles
            self._load_policy_no_lock()
            # Avoid calling get_user_roles, as it will require the lock.
            enforcer = self._ensure_enforcer()
            current_roles = enforcer.get_roles_for_user(user_id)
            if not current_roles:
                logger.debug(f'User {user_id} has no roles')
                return
            enforcer.remove_grouping_policy(user_id, current_roles[0])
            enforcer.save_policy()

    def update_role(self, user_id: str, new_role: str) -> None:
        """Update user role relationship."""
        with _policy_lock():
            # Get current roles
            self._load_policy_no_lock()
            # Avoid calling get_user_roles, as it will require the lock.
            enforcer = self._ensure_enforcer()
            current_roles = enforcer.get_roles_for_user(user_id)
            if not current_roles:
                logger.debug(f'User {user_id} has no roles')
            else:
                # TODO(hailong): how to handle multiple roles?
                current_role = current_roles[0]
                if current_role == new_role:
                    logger.debug(f'User {user_id} already has role {new_role}')
                    return
                enforcer.remove_grouping_policy(user_id, current_role)

            # Update user role
            enforcer.add_grouping_policy(user_id, new_role)
            enforcer.save_policy()

    def get_user_roles(self, user_id: str) -> List[str]:
        """Get all roles for a user.

        This method returns all roles that the user has, including inherited
        roles. For example, if a user has role 'admin' and 'admin' inherits
        from 'user', this method will return ['admin', 'user'].

        Args:
            user: The user ID to get roles for.

        Returns:
            A list of role names that the user has.
        """
        self._load_policy_no_lock()
        enforcer = self._ensure_enforcer()
        return enforcer.get_roles_for_user(user_id)

    def get_users_for_role(self, role: str) -> List[str]:
        """Get all users for a role."""
        self._load_policy_no_lock()
        enforcer = self._ensure_enforcer()
        return enforcer.get_users_for_role(role)

    def check_endpoint_permission(self, user_id: str, path: str,
                                  method: str) -> bool:
        """Check permission."""
        # We intentionally don't load the policy here, as it is a hot path, and
        # we don't support updating the policy.
        # We don't hold the lock for checking permission, as it is read only and
        # it is a hot path in every request. It is ok to have a stale policy,
        # as long as it is eventually consistent.
        # self._load_policy_no_lock()
        enforcer = self._ensure_enforcer()
        return enforcer.enforce(user_id, path, method)

    def _load_policy_no_lock(self):
        """Load policy from storage."""
        enforcer = self._ensure_enforcer()
        enforcer.load_policy()

    def load_policy(self):
        """Load policy from storage with lock."""
        with _policy_lock():
            self._load_policy_no_lock()

    # Right now, not a lot of users are using multiple workspaces,
    # so 5 should be more than enough.
    @annotations.lru_cache(scope='request', maxsize=5)
    def check_workspace_permission(self, user_id: str,
                                   workspace_name: str) -> bool:
        """Check workspace permission.

        This method checks if a user has permission to access a specific
        workspace.

        For private workspaces, the user must have explicit permission.

        For public workspaces, the permission is granted via a wildcard policy
        ('*').
        """
        if os.getenv(constants.ENV_VAR_IS_SKYPILOT_SERVER) is None:
            # When it is not on API server, we allow all users to access all
            # workspaces, as the workspace check has been done on API server.
            return True
        role = self.get_user_roles(user_id)
        if rbac.RoleName.ADMIN.value in role:
            return True
        # The Casbin model matcher already handles the wildcard '*' case:
        # m = (g(r.sub, p.sub)|| p.sub == '*') && r.obj == p.obj &&
        # r.act == p.act
        # This means if there's a policy ('*', workspace_name, '*'), it will
        # match any user
        enforcer = self._ensure_enforcer()
        result = enforcer.enforce(user_id, workspace_name, '*')
        logger.debug(f'Workspace permission check: user={user_id}, '
                     f'workspace={workspace_name}, result={result}')
        return result

    def check_service_account_token_permission(self, user_id: str,
                                               token_owner_id: str,
                                               action: str) -> bool:
        """Check service account token permission.

        This method checks if a user has permission to perform an action on
        a service account token owned by another user.

        Args:
            user_id: The ID of the user requesting the action
            token_owner_id: The ID of the user who owns the token
            action: The action being performed (e.g., 'delete', 'view')

        Returns:
            True if the user has permission, False otherwise
        """
        del action
        # Users can always manage their own tokens
        if user_id == token_owner_id:
            return True

        # Check if user has admin role (admins can manage any token)
        user_roles = self.get_user_roles(user_id)
        if rbac.RoleName.ADMIN.value in user_roles:
            return True

        # Regular users cannot manage tokens owned by others
        return False

    def add_workspace_policy(self, workspace_name: str,
                             users: List[str]) -> None:
        """Add workspace policy.

        Args:
            workspace_name: Name of the workspace
            users: List of user IDs that should have access.
                   For public workspaces, this should be ['*'].
                   For private workspaces, this should be specific user IDs.
        """
        with _policy_lock():
            enforcer = self._ensure_enforcer()
            for user in users:
                logger.debug(f'Adding workspace policy: user={user}, '
                             f'workspace={workspace_name}')
                enforcer.add_policy(user, workspace_name, '*')
            enforcer.save_policy()

    def update_workspace_policy(self, workspace_name: str,
                                users: List[str]) -> None:
        """Update workspace policy.

        Args:
            workspace_name: Name of the workspace
            users: List of user IDs that should have access.
                   For public workspaces, this should be ['*'].
                   For private workspaces, this should be specific user IDs.
        """
        with _policy_lock():
            self._load_policy_no_lock()
            enforcer = self._ensure_enforcer()
            # Remove all existing policies for this workspace
            enforcer.remove_filtered_policy(1, workspace_name)
            # Add new policies
            for user in users:
                logger.debug(f'Updating workspace policy: user={user}, '
                             f'workspace={workspace_name}')
                enforcer.add_policy(user, workspace_name, '*')
            enforcer.save_policy()

    def remove_workspace_policy(self, workspace_name: str) -> None:
        """Remove workspace policy."""
        with _policy_lock():
            enforcer = self._ensure_enforcer()
            enforcer.remove_filtered_policy(1, workspace_name)
            enforcer.save_policy()


@contextlib.contextmanager
def _policy_lock() -> Generator[None, None, None]:
    """Context manager for policy update lock."""
    try:
        with filelock.FileLock(POLICY_UPDATE_LOCK_PATH,
                               POLICY_UPDATE_LOCK_TIMEOUT_SECONDS):
            yield
    except filelock.Timeout as e:
        raise RuntimeError(f'Failed to reload policy due to a timeout '
                           f'when trying to acquire the lock at '
                           f'{POLICY_UPDATE_LOCK_PATH}. '
                           'Please try again or manually remove the lock '
                           f'file if you believe it is stale.') from e


# Singleton instance of PermissionService for other modules to use.
permission_service = PermissionService()



"""Utilities for Kubernetes ConfigMap operations in SkyPilot."""
import os

from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import kubernetes
from sky.utils import yaml_utils

logger = sky_logging.init_logger(__name__)

# Kubernetes ConfigMap sync constants
_KUBE_SERVICE_ACCOUNT_PATH = '/var/run/secrets/kubernetes.io/serviceaccount'
_CONFIGMAP_SYNC_TIMEOUT = 10  # seconds


def is_running_in_kubernetes() -> bool:
    """Check if we're running inside a Kubernetes pod."""
    return os.path.exists(f'{_KUBE_SERVICE_ACCOUNT_PATH}/token')


def _get_kubernetes_namespace() -> str:
    """Get the current Kubernetes namespace from the service account."""
    try:
        namespace_file = f'{_KUBE_SERVICE_ACCOUNT_PATH}/namespace'
        if os.path.exists(namespace_file):
            with open(namespace_file, encoding='utf-8') as f:
                return f.read().strip()
    except (OSError, IOError):
        pass
    return 'default'


def _get_configmap_name() -> str:
    """Get the ConfigMap name for the SkyPilot config."""
    release_name = (os.getenv('HELM_RELEASE_NAME') or
                    os.getenv('SKYPILOT_RELEASE_NAME') or 'skypilot')
    return f'{release_name}-config'


def initialize_configmap_sync_on_startup(config_file_path: str) -> None:
    """Initialize ConfigMap sync on API server startup.

    This syncs existing PVC config to ConfigMap if ConfigMap doesn't exist.
    This handles the upgrade scenario where an existing deployment has
    workspace configs on PVC but no ConfigMap exists.

    Args:
        config_file_path: Path to the config file to sync.
    """
    config_file_path = os.path.expanduser(config_file_path)
    if not is_running_in_kubernetes() or not os.path.exists(config_file_path):
        return

    try:
        namespace = _get_kubernetes_namespace()
        configmap_name = _get_configmap_name()

        # Check if ConfigMap exists
        try:
            kubernetes.core_api().read_namespaced_config_map(
                name=configmap_name, namespace=namespace)
            # ConfigMap exists, don't overwrite it
            logger.debug(f'ConfigMap {configmap_name} already exists')
            return
        except kubernetes.kubernetes.client.rest.ApiException as e:
            if e.status != 404:
                raise
            # ConfigMap doesn't exist, create it

        current_config = skypilot_config.parse_and_validate_config_file(
            config_file_path)
        config_yaml = yaml_utils.dump_yaml_str(dict(current_config))

        configmap_body = {
            'apiVersion': 'v1',
            'kind': 'ConfigMap',
            'metadata': {
                'name': configmap_name,
                'namespace': namespace,
                'labels': {
                    'app.kubernetes.io/name': 'skypilot',
                    'app.kubernetes.io/component': 'config'
                }
            },
            'data': {
                'config.yaml': config_yaml
            }
        }

        kubernetes.core_api().create_namespaced_config_map(
            namespace=namespace,
            body=configmap_body,
            _request_timeout=_CONFIGMAP_SYNC_TIMEOUT)

        logger.info(f'Synced PVC config to new ConfigMap {configmap_name}')

    except Exception as e:  # pylint: disable=broad-except
        logger.warning(f'Failed to initialize ConfigMap sync: {e}')


def patch_configmap_with_config(config, config_file_path: str) -> None:
    """Patch the Kubernetes ConfigMap with the updated config.

    Args:
        config: The updated config to sync to the ConfigMap.
        config_file_path: Path to the config file for fallback sync.
    """
    if not is_running_in_kubernetes():
        return

    try:
        namespace = _get_kubernetes_namespace()
        configmap_name = _get_configmap_name()
        config_yaml = yaml_utils.dump_yaml_str(dict(config))
        patch_body = {'data': {'config.yaml': config_yaml}}

        try:
            kubernetes.core_api().patch_namespaced_config_map(
                name=configmap_name,
                namespace=namespace,
                body=patch_body,
                _request_timeout=_CONFIGMAP_SYNC_TIMEOUT)
            logger.debug(f'Synced config to ConfigMap {configmap_name}')
        except kubernetes.kubernetes.client.rest.ApiException as e:
            if e.status == 404:
                # ConfigMap doesn't exist, create it
                logger.info(f'ConfigMap {configmap_name} not found, creating')
                initialize_configmap_sync_on_startup(config_file_path)
            else:
                raise

    except Exception as e:  # pylint: disable=broad-except
        logger.warning(f'Failed to sync config to ConfigMap: {e}')



"""Utility functions for generating instance links for cloud providers."""
from typing import Dict

from sky import sky_logging
from sky.provision import common
from sky.provision import constants as provision_constants

logger = sky_logging.init_logger(__name__)

# URL templates for each cloud provider
# Placeholders:
#   {region} - Cloud region
#   {project_id} - GCP project ID
#   {subscription_id} - Azure subscription ID
#   {resource_group} - Azure resource group
#   {tag_key} - Tag key used to identify cluster instances
#   {cluster_name} - Name of the cluster

AWS_INSTANCES_URL = ('https://{region}.console.aws.amazon.com/ec2/v2/home'
                     '?region={region}#Instances:tag:{tag_key}={cluster_name}')

# Azure doesn't support direct tag filter URLs, so we link to the resource group
AZURE_RESOURCE_GROUP_URL = (
    'https://portal.azure.com/#@/resource/subscriptions'
    '/{subscription_id}/resourceGroups/{resource_group}/overview')

# GCP Console base URL
GCP_INSTANCES_BASE_URL = 'https://console.cloud.google.com/compute/instances'


def _build_gcp_instances_url(project_id: str, tag_key: str,
                             cluster_name: str) -> str:
    """Build GCP instances URL with label filter.

    GCP Console uses a pageState parameter with a specially encoded filter.
    The filter JSON structure is:
        [{"k":"","t":10,"v":"\"label_key:label_value\"","s":true}]

    Where:
        - k: filter key (empty for label filters)
        - t: filter type (10 = label filter)
        - v: filter value with escaped quotes around "label_key:label_value"
        - s: unknown, always true

    GCP uses a mix of:
        - Standard URL encoding for outer structure (%22 for ")
        - Underscore notation inside the filter (_22 for ", _3A for :, etc.)
        - Double URL-encoding for brackets (%255B = %5B = [)
    """
    # Build the filter value: \"tag_key:cluster_name\"
    # Using underscore notation: _5C_22 = \", _3A = :
    filter_value = f'_5C_22{tag_key}_3A{cluster_name}_5C_22'

    # Build the filter object using underscore notation for internal quotes and
    # colons.
    # {"k":"","t":10,"v":"<filter_value>","s":true}
    # _22 = ", _3A = :, _2C = ,
    filter_obj = (
        f'_22k_22_3A_22_22_2C'  # "k":"",
        f'_22t_22_3A10_2C'  # "t":10,
        f'_22v_22_3A_22{filter_value}_22_2C'  # "v":"<value>",
        f'_22s_22_3Atrue')  # "s":true

    # Wrap in array brackets (double URL-encoded: %255B = %5B = [, %257D = %7D)
    filter_array = f'%255B%257B{filter_obj}%257D%255D'

    # Build pageState: ("instances":("p":0,"f":"<filter>"))
    # %22 = " (standard URL encoding)
    page_state = f'(%22instances%22:(%22p%22:0,%22f%22:%22{filter_array}%22))'

    return (
        f'{GCP_INSTANCES_BASE_URL}?project={project_id}&pageState={page_state}')


def generate_instance_links(
    cluster_info: common.ClusterInfo,
    cluster_name: str,
) -> Dict[str, str]:
    """Generate instance links for a cluster based on the cloud provider.

    Creates links to filtered views in cloud consoles that show all instances
    belonging to the cluster (useful for multi-node jobs).

    Args:
        cluster_info: ClusterInfo object containing instance information.
        cluster_name: Cluster name for tag-based filtering.

    Returns:
        Dictionary mapping link labels to URLs. Empty dict if links cannot be
        generated (e.g., for Kubernetes or unsupported clouds).
    """
    links: Dict[str, str] = {}
    provider_name = cluster_info.provider_name.lower()
    provider_config = cluster_info.provider_config or {}

    # Skip Kubernetes and other non-cloud providers
    if provider_name in ('kubernetes', 'local'):
        return links

    # Tag used by SkyPilot to identify cluster instances
    tag_key = provision_constants.TAG_RAY_CLUSTER_NAME

    if provider_name == 'aws':
        region = provider_config.get('region')
        if not region:
            logger.debug('AWS region not found in provider config, '
                         'skipping instance links')
            return links
        links['AWS Instances'] = AWS_INSTANCES_URL.format(
            region=region,
            tag_key=tag_key,
            cluster_name=cluster_name,
        )

    elif provider_name == 'gcp':
        project_id = provider_config.get('project_id')
        if not project_id:
            logger.debug('GCP project_id not found in provider config, '
                         'skipping instance links')
            return links
        links['GCP Instances'] = _build_gcp_instances_url(
            project_id=project_id,
            tag_key=tag_key,
            cluster_name=cluster_name,
        )

    elif provider_name == 'azure':
        subscription_id = provider_config.get('subscription_id')
        resource_group = provider_config.get('resource_group')
        if not subscription_id or not resource_group:
            logger.debug('Azure subscription_id or resource_group not found '
                         'in provider config, skipping instance links')
            return links
        links['Azure Resource Group'] = AZURE_RESOURCE_GROUP_URL.format(
            subscription_id=subscription_id,
            resource_group=resource_group,
        )

    return links



"""PrimeIntellect service catalog.

This module loads the service catalog file and can be used to
query instance types and pricing information for PrimeIntellect.
"""

import typing
from typing import Dict, List, Optional, Tuple, Union

from sky.catalog import common

if typing.TYPE_CHECKING:
    from sky.clouds import cloud

_df = common.read_catalog('primeintellect/vms.csv')


def instance_type_exists(instance_type: str) -> bool:
    return common.instance_type_exists_impl(_df, instance_type)


def validate_region_zone(
        region: Optional[str],
        zone: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
    return common.validate_region_zone_impl('primeintellect', _df, region, zone)


def get_hourly_cost(instance_type: str,
                    use_spot: bool = False,
                    region: Optional[str] = None,
                    zone: Optional[str] = None) -> float:
    """Returns the cost, or the cheapest cost among all zones for spot."""
    return common.get_hourly_cost_impl(_df, instance_type, use_spot, region,
                                       zone)


def get_vcpus_mem_from_instance_type(
        instance_type: str) -> Tuple[Optional[float], Optional[float]]:
    return common.get_vcpus_mem_from_instance_type_impl(_df, instance_type)


def get_default_instance_type(cpus: Optional[str] = None,
                              memory: Optional[str] = None,
                              disk_tier: Optional[str] = None,
                              local_disk: Optional[str] = None,
                              region: Optional[str] = None,
                              zone: Optional[str] = None) -> Optional[str]:
    del disk_tier, local_disk  # no disk tiers
    return common.get_instance_type_for_cpus_mem_impl(_df, cpus, memory, region,
                                                      zone)


def get_accelerators_from_instance_type(
        instance_type: str) -> Optional[Dict[str, Union[int, float]]]:
    return common.get_accelerators_from_instance_type_impl(_df, instance_type)


def get_instance_type_for_accelerator(
        acc_name: str,
        acc_count: int,
        cpus: Optional[str] = None,
        memory: Optional[str] = None,
        use_spot: bool = False,
        local_disk: Optional[str] = None,
        region: Optional[str] = None,
        zone: Optional[str] = None) -> Tuple[Optional[List[str]], List[str]]:
    """Returns a list of instance types that have the given accelerator."""
    del local_disk  # unused
    return common.get_instance_type_for_accelerator_impl(df=_df,
                                                         acc_name=acc_name,
                                                         acc_count=acc_count,
                                                         cpus=cpus,
                                                         memory=memory,
                                                         use_spot=use_spot,
                                                         region=region,
                                                         zone=zone)


def get_region_zones_for_instance_type(instance_type: str,
                                       use_spot: bool) -> List['cloud.Region']:
    df = _df[_df['InstanceType'] == instance_type]
    return common.get_region_zones(df, use_spot)


def list_accelerators(
        gpus_only: bool,
        name_filter: Optional[str],
        region_filter: Optional[str],
        quantity_filter: Optional[int],
        case_sensitive: bool = True,
        all_regions: bool = False,
        require_price: bool = True) -> Dict[str, List[common.InstanceTypeInfo]]:
    """Returns all instance types in Prime Intellect offering GPUs."""
    del require_price
    return common.list_accelerators_impl('PrimeIntellect', _df, gpus_only,
                                         name_filter, region_filter,
                                         quantity_filter, case_sensitive,
                                         all_regions)



"""REST API for Recipe Hub.

This module provides the FastAPI router for recipe management,
including CRUD operations, pinning, and deployment.
"""
import fastapi

from sky.recipes import core
from sky.server.requests import executor
from sky.server.requests import payloads
from sky.server.requests import request_names
from sky.server.requests import requests as api_requests

router = fastapi.APIRouter()


@router.get('')
async def list_recipes(request: fastapi.Request) -> None:
    """List recipes (pinned + user's own)."""
    auth_user = request.state.auth_user
    auth_user_env_vars_kwargs = {
        'env_vars': auth_user.to_env_vars()
    } if auth_user else {}
    request_body = payloads.RecipeListBody(**auth_user_env_vars_kwargs)

    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_LIST,
        request_body=request_body,
        func=core.list_recipes,
        schedule_type=api_requests.ScheduleType.SHORT,
    )


@router.post('/list')
async def list_recipes_with_filters(
    request: fastapi.Request,
    list_body: payloads.RecipeListBody,
) -> None:
    """List recipes with filters."""
    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_LIST,
        request_body=list_body,
        func=core.list_recipes,
        schedule_type=api_requests.ScheduleType.SHORT,
    )


@router.post('/get')
async def get_recipe(
    request: fastapi.Request,
    get_body: payloads.RecipeGetBody,
) -> None:
    """Get a single recipe by name."""
    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_GET,
        request_body=get_body,
        func=core.get_recipe,
        schedule_type=api_requests.ScheduleType.SHORT,
    )


@router.post('/create')
async def create_recipe(
    request: fastapi.Request,
    create_body: payloads.RecipeCreateBody,
) -> None:
    """Create a new recipe."""
    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_CREATE,
        request_body=create_body,
        func=core.create_recipe,
        schedule_type=api_requests.ScheduleType.SHORT,
    )


@router.post('/update')
async def update_recipe(
    request: fastapi.Request,
    update_body: payloads.RecipeUpdateBody,
) -> None:
    """Update an existing recipe."""
    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_UPDATE,
        request_body=update_body,
        func=core.update_recipe,
        schedule_type=api_requests.ScheduleType.SHORT,
    )


@router.post('/delete')
async def delete_recipe(
    request: fastapi.Request,
    delete_body: payloads.RecipeDeleteBody,
) -> None:
    """Delete a recipe."""
    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_DELETE,
        request_body=delete_body,
        func=core.delete_recipe,
        schedule_type=api_requests.ScheduleType.SHORT,
    )


@router.post('/pin')
async def pin_recipe(
    request: fastapi.Request,
    pin_body: payloads.RecipePinBody,
) -> None:
    """Toggle pin status of a recipe (admin only)."""
    # Note: Admin check should be performed in the core function or
    # via middleware. For now, the API is available to all authenticated users.
    await executor.schedule_request_async(
        request_id=request.state.request_id,
        request_name=request_names.RequestName.RECIPE_PIN,
        request_body=pin_body,
        func=core.toggle_pin,
        schedule_type=api_requests.ScheduleType.SHORT,
    )



"""Kubernetes pvc provisioning."""
from typing import Any, Dict, List, Optional, Set, Tuple

from sky import global_user_state
from sky import models
from sky import sky_logging
from sky.adaptors import kubernetes
from sky.provision import constants
from sky.provision.kubernetes import config as config_lib
from sky.provision.kubernetes import constants as k8s_constants
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.utils import resources_utils
from sky.utils import volume as volume_lib

logger = sky_logging.init_logger(__name__)

PVC_FAILING_EVENT_REASONS = ('ProvisioningFailed',)
WARNING_EVENT_TYPE = 'Warning'


def _get_context_namespace(config: models.VolumeConfig) -> Tuple[str, str]:
    """Gets the context and namespace of a volume."""
    if config.region is None:
        context = kubernetes_utils.get_current_kube_config_context_name()
        config.region = context
    else:
        context = config.region
    namespace = config.config.get('namespace')
    if namespace is None:
        namespace = kubernetes_utils.get_kube_config_context_namespace(context)
        config.config['namespace'] = namespace
    return context, namespace


def check_pvc_usage_for_pod(context: Optional[str], namespace: str,
                            pod_spec: Dict[str, Any]) -> None:
    """Checks if the PVC is used by any pod in the namespace."""
    volumes = pod_spec.get('spec', {}).get('volumes', [])
    if not volumes:
        return
    once_modes = [
        volume_lib.VolumeAccessMode.READ_WRITE_ONCE.value,
        volume_lib.VolumeAccessMode.READ_WRITE_ONCE_POD.value
    ]
    for volume in volumes:
        pvc_name = volume.get('persistentVolumeClaim', {}).get('claimName')
        if not pvc_name:
            continue
        pvc = kubernetes.core_api(
            context).read_namespaced_persistent_volume_claim(
                name=pvc_name,
                namespace=namespace,
                _request_timeout=kubernetes.API_TIMEOUT)
        access_mode = pvc.spec.access_modes[0]
        if access_mode not in once_modes:
            continue
        usedby_pods, _ = _get_volume_usedby(context, namespace, pvc_name)
        if usedby_pods:
            raise config_lib.KubernetesError(f'Volume {pvc_name} with access '
                                             f'mode {access_mode} is already '
                                             f'in use by Pods {usedby_pods}.')


def apply_volume(config: models.VolumeConfig) -> models.VolumeConfig:
    """Creates or registers a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_spec = _get_pvc_spec(namespace, config)
    # Check if the storage class exists
    storage_class_name = pvc_spec['spec'].get('storageClassName')
    if storage_class_name is not None:
        try:
            kubernetes.storage_api(context).read_storage_class(
                name=storage_class_name,
                _request_timeout=kubernetes.API_TIMEOUT)
        except kubernetes.api_exception() as e:
            raise config_lib.KubernetesError(
                f'Check storage class {storage_class_name} error: {e}')
    create_persistent_volume_claim(namespace, context, pvc_spec, config)
    return config


def delete_volume(config: models.VolumeConfig) -> models.VolumeConfig:
    """Deletes a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_name = config.name_on_cloud
    kubernetes_utils.delete_k8s_resource_with_retry(
        delete_func=lambda pvc_name=pvc_name: kubernetes.core_api(
            context).delete_namespaced_persistent_volume_claim(
                name=pvc_name,
                namespace=namespace,
                _request_timeout=kubernetes.API_TIMEOUT),
        resource_type='pvc',
        resource_name=pvc_name)
    logger.info(f'Deleted PVC {pvc_name} in namespace {namespace}')
    return config


def _get_volume_usedby(
    context: Optional[str],
    namespace: str,
    pvc_name: str,
) -> Tuple[List[str], List[str]]:
    """Gets the usedby resources of a volume.

    This function returns the pods and clusters that are using the volume.
    The usedby_pods is accurate, which also includes the Pods that are not
    managed by SkyPilot.

    Args:
        context: Kubernetes context
        namespace: Kubernetes namespace
        pvc_name: PVC name

    Returns:
        usedby_pods: List of pods using the volume. These may include pods
                     not created by SkyPilot.
        usedby_clusters: List of clusters using the volume.
    """
    usedby_pods = []
    usedby_clusters = []
    field_selector = ','.join([
        f'status.phase!={phase}'
        for phase in k8s_constants.PVC_NOT_HOLD_POD_PHASES
    ])
    cloud_to_name_map = _get_cluster_name_on_cloud_to_cluster_name_map()
    # Get all pods in the namespace
    pods = kubernetes.core_api(context).list_namespaced_pod(
        namespace=namespace,
        field_selector=field_selector,
        _request_timeout=kubernetes.API_TIMEOUT)
    for pod in pods.items:
        if pod.spec.volumes is None:
            continue
        for volume in pod.spec.volumes:
            if volume.persistent_volume_claim is None:
                continue
            if volume.persistent_volume_claim.claim_name == pvc_name:
                usedby_pods.append(pod.metadata.name)
                # Get the real cluster name
                cluster_name_on_cloud = pod.metadata.labels.get(
                    constants.TAG_SKYPILOT_CLUSTER_NAME)
                if cluster_name_on_cloud is None:
                    continue
                cluster_name = cloud_to_name_map.get(cluster_name_on_cloud)
                if cluster_name is not None:
                    usedby_clusters.append(cluster_name)
    if usedby_pods:
        logger.debug(f'Volume {pvc_name} is used by Pods {usedby_pods}'
                     f' and clusters {usedby_clusters}')
    return usedby_pods, usedby_clusters


def _get_cluster_name_on_cloud_to_cluster_name_map() -> Dict[str, str]:
    """Gets the map from cluster name on cloud to cluster name."""
    clusters = global_user_state.get_clusters()
    cloud_to_name_map = {}
    for cluster in clusters:
        handle = cluster['handle']
        if handle is None:
            continue
        cloud_to_name_map[handle.cluster_name_on_cloud] = cluster['name']
    return cloud_to_name_map


def get_volume_usedby(
    config: models.VolumeConfig,) -> Tuple[List[str], List[str]]:
    """Gets the usedby resources of a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_name = config.name_on_cloud
    return _get_volume_usedby(context, namespace, pvc_name)


def refresh_volume_config(
    config: models.VolumeConfig,) -> Tuple[bool, models.VolumeConfig]:
    """Refreshes the volume config.
    For volume config with region None, we need to set the region to the
    in-cluster context name.

    Returns:
        need_refresh: Whether need to refresh the volume config.
        volume_config: The volume config to be refreshed.
    """
    if config.region is None:
        config.region = kubernetes.in_cluster_context_name()
        return True, config
    return False, config


def get_all_volumes_usedby(
    configs: List[models.VolumeConfig],
) -> Tuple[Dict[str, Any], Dict[str, Any], Set[str]]:
    """Gets the usedby resources of all volumes.

    Args:
        configs: List of VolumeConfig objects.

    Returns:
        usedby_pods: Dictionary of context to namespace to volume name to pods
                     using the volume. These may include pods not created by
                     SkyPilot.
        usedby_clusters: Dictionary of context to namespace to volume name to
                         clusters using the volume.
        failed_volume_names: Set of volume names whose usedby info failed to
          fetch.
    """
    field_selector = ','.join([
        f'status.phase!={phase}'
        for phase in k8s_constants.PVC_NOT_HOLD_POD_PHASES
    ])
    label_selector = 'parent=skypilot'
    context_to_namespaces: Dict[str, Set[str]] = {}
    pvc_names = set()
    original_volume_names: Dict[str, Dict[str, List[str]]] = {}
    for config in configs:
        context, namespace = _get_context_namespace(config)
        context_to_namespaces.setdefault(context, set()).add(namespace)
        original_volume_names.setdefault(context,
                                         {}).setdefault(namespace,
                                                        []).append(config.name)
        pvc_names.add(config.name_on_cloud)
    cloud_to_name_map = _get_cluster_name_on_cloud_to_cluster_name_map()
    # Get all pods in the namespace
    used_by_pods: Dict[str, Dict[str, Dict[str, List[str]]]] = {}
    used_by_clusters: Dict[str, Dict[str, Dict[str, List[str]]]] = {}
    failed_volume_names: Set[str] = set()
    for context, namespaces in context_to_namespaces.items():
        used_by_pods[context] = {}
        used_by_clusters[context] = {}
        for namespace in namespaces:
            used_by_pods[context][namespace] = {}
            used_by_clusters[context][namespace] = {}
            try:
                pods = kubernetes.core_api(context).list_namespaced_pod(
                    namespace=namespace,
                    field_selector=field_selector,
                    label_selector=label_selector,
                    _request_timeout=kubernetes.API_TIMEOUT)
            except Exception as e:  # pylint: disable=broad-except
                logger.debug(f'Failed to get pods in namespace {namespace} '
                             f'in context {context}: {e}')
                # Mark all volumes in this namespace as failed
                for original_volume_name in original_volume_names[context][
                        namespace]:
                    failed_volume_names.add(original_volume_name)
                continue
            for pod in pods.items:
                if pod.spec.volumes is None:
                    continue
                for volume in pod.spec.volumes:
                    if volume.persistent_volume_claim is None:
                        continue
                    volume_name = volume.persistent_volume_claim.claim_name
                    if volume_name not in pvc_names:
                        continue
                    if volume_name not in used_by_pods[context][namespace]:
                        used_by_pods[context][namespace][volume_name] = []
                    used_by_pods[context][namespace][volume_name].append(
                        pod.metadata.name)
                    cluster_name_on_cloud = pod.metadata.labels.get(
                        constants.TAG_SKYPILOT_CLUSTER_NAME)
                    if cluster_name_on_cloud is None:
                        continue
                    cluster_name = cloud_to_name_map.get(cluster_name_on_cloud)
                    if cluster_name is None:
                        continue
                    if cluster_name not in used_by_clusters[context][namespace]:
                        used_by_clusters[context][namespace][cluster_name] = []
                    used_by_clusters[context][namespace][cluster_name].append(
                        cluster_name)
    return used_by_pods, used_by_clusters, failed_volume_names


def map_all_volumes_usedby(
        used_by_pods: Dict[str, Any], used_by_clusters: Dict[str, Any],
        config: models.VolumeConfig) -> Tuple[List[str], List[str]]:
    """Maps the usedby resources of a volume."""
    context, namespace = _get_context_namespace(config)
    pvc_name = config.name_on_cloud

    return (used_by_pods.get(context, {}).get(namespace, {}).get(pvc_name, []),
            used_by_clusters.get(context, {}).get(namespace,
                                                  {}).get(pvc_name, []))


def get_all_volumes_errors(
    configs: List[models.VolumeConfig],) -> Dict[str, Optional[str]]:
    """Gets error messages for all Kubernetes PVC volumes.

    Checks if PVCs are in Pending state and if so, checks for access mode
    mismatches between the PVC and the storage class's allowed access modes.

    Args:
        configs: List of VolumeConfig objects.

    Returns:
        Dictionary mapping volume name to error message (None if no error).
    """
    context_to_namespaces: Dict[str, Set[str]] = {}
    config_by_pvc_name: Dict[str, Dict[str, models.VolumeConfig]] = {}

    for config in configs:
        context, namespace = _get_context_namespace(config)
        context_to_namespaces.setdefault(context, set()).add(namespace)
        config_by_pvc_name.setdefault(context,
                                      {})[config.name_on_cloud] = config

    volume_errors: Dict[str, Optional[str]] = {}

    for context, namespaces in context_to_namespaces.items():
        for namespace in namespaces:
            try:
                # List all PVCs in the namespace with the skypilot label
                pvcs = kubernetes.core_api(
                    context).list_namespaced_persistent_volume_claim(
                        namespace=namespace,
                        label_selector='parent=skypilot',
                        _request_timeout=kubernetes.API_TIMEOUT)
            except Exception as e:  # pylint: disable=broad-except
                logger.debug(f'Failed to get PVCs in namespace {namespace} '
                             f'in context {context}: {e}')
                continue

            for pvc in pvcs.items:
                pvc_name = pvc.metadata.name
                vol_config = config_by_pvc_name.get(context, {}).get(pvc_name)
                if vol_config is None:
                    continue

                volume_name = vol_config.name
                pvc_phase = pvc.status.phase

                # If PVC is bound, no error
                if pvc_phase == 'Bound':
                    volume_errors[volume_name] = None
                    continue

                # If PVC is pending, check for access mode mismatch
                if pvc_phase == 'Pending':
                    error_msg = _check_pvc_access_mode_error(context, pvc)
                    if error_msg:
                        volume_errors[volume_name] = error_msg
                    else:
                        volume_binding_mode = (
                            _check_storage_class_volume_binding_mode(
                                context, pvc))
                        if (volume_binding_mode is None or
                                volume_binding_mode == 'WaitForFirstConsumer'):
                            error_msg = None
                            pvc_events = kubernetes_utils.get_pvc_events(
                                context, namespace, pvc_name)
                            for event in pvc_events:
                                if (event.type == WARNING_EVENT_TYPE or
                                        event.reason
                                        in PVC_FAILING_EVENT_REASONS):
                                    reason_str = event.reason
                                    if event.message:
                                        reason_str += f': {event.message}'
                                    error_msg = (f'PVC is pending. '
                                                 f'{reason_str}. To debug, run'
                                                 f': kubectl describe pvc '
                                                 f'{pvc_name} -n {namespace}')
                                    break
                            volume_errors[volume_name] = error_msg
                        else:
                            # Generic pending message if no specific error
                            # detected
                            volume_errors[volume_name] = (
                                'PVC is pending. This may be due to '
                                'insufficient storage resources or '
                                'misconfiguration. To debug, run: '
                                f'kubectl describe pvc {pvc_name} -n '
                                f'{namespace}')
                elif pvc_phase == 'Lost':
                    volume_errors[volume_name] = (
                        'PVC is in Lost state. The bound PersistentVolume '
                        'has been deleted or is unavailable. To debug, '
                        f'run: kubectl describe pvc {pvc_name} -n {namespace}')
                else:
                    # Other phases (e.g., Terminating)
                    volume_errors[volume_name] = None

    return volume_errors


def _check_storage_class_volume_binding_mode(context: Optional[str],
                                             pvc: Any) -> Optional[str]:
    """Check the volumeBindingMode of the storage class for the PVC.

    Args:
        context: Kubernetes context
        pvc: V1PersistentVolumeClaim object

    Returns:
        volumeBindingMode of the storage class for the PVC,
        None if failed to read the storage class.
    """
    storage_class_name = pvc.spec.storage_class_name
    if not storage_class_name:
        return None
    try:
        storage_class = kubernetes.storage_api(context).read_storage_class(
            name=storage_class_name, _request_timeout=kubernetes.API_TIMEOUT)
        return storage_class.volume_binding_mode
    except Exception as e:  # pylint: disable=broad-except
        logger.debug(f'Failed to read storage class {storage_class_name}: {e}')
        return None


def _check_pvc_access_mode_error(context: Optional[str],
                                 pvc: Any) -> Optional[str]:
    """Check if a pending PVC has an access mode mismatch.

    Args:
        context: Kubernetes context
        pvc: V1PersistentVolumeClaim object

    Returns:
        Error message if there's an access mode mismatch, None otherwise.
    """
    pvc_access_modes = pvc.spec.access_modes or []
    if not pvc_access_modes:
        return None

    pvc_access_mode = pvc_access_modes[0]
    storage_class_name = pvc.spec.storage_class_name

    # Try to find available PVs and check their access modes
    try:
        pvs = kubernetes.core_api(context).list_persistent_volume(
            _request_timeout=kubernetes.API_TIMEOUT)
    except Exception as e:  # pylint: disable=broad-except
        logger.debug(f'Failed to list PVs: {e}')
        return None

    # Filter PVs that match the storage class and are available
    available_pvs = []
    for pv in pvs.items:
        # Check if PV matches storage class
        pv_storage_class = pv.spec.storage_class_name
        if storage_class_name and pv_storage_class != storage_class_name:
            continue
        # Check if PV is available
        if pv.status.phase == 'Available':
            available_pvs.append(pv)

    if not available_pvs:
        return None

    # Check if any available PV has a compatible access mode
    for pv in available_pvs:
        pv_access_modes = pv.spec.access_modes or []
        if pvc_access_mode in pv_access_modes:
            # Found a compatible PV, so access mode is not the issue
            return None

    # No compatible PV found - access mode mismatch
    pv_access_modes_str = ', '.join(
        sorted(
            set(mode for pv in available_pvs
                for mode in (pv.spec.access_modes or []))))
    pvc_name = pvc.metadata.name
    namespace = pvc.metadata.namespace
    return (f'PVC access mode mismatch: PVC requests {pvc_access_mode}, but '
            f'available PersistentVolumes support: {pv_access_modes_str}. '
            f'Update the volume with the correct access_mode '
            f'(e.g., {pv_access_modes_str}) and recreate it. To debug, run: '
            f'kubectl describe pvc {pvc_name} -n {namespace}')


def _populate_config_from_pvc(config: models.VolumeConfig,
                              pvc_obj: Any) -> None:
    """Populate missing fields in config from a PVC object.

    Args:
        config: VolumeConfig to populate
        pvc_obj: V1PersistentVolumeClaim object from kubernetes client
    """
    if pvc_obj is None:
        return
    pvc_name = pvc_obj.metadata.name

    # Populate storageClassName if not set
    if config.config.get('storage_class_name') is None:
        pvc_storage_class = getattr(pvc_obj.spec, 'storage_class_name', None)
        if pvc_storage_class:
            config.config['storage_class_name'] = pvc_storage_class

    # Populate size if not set (prefer bound capacity, fallback to requested)
    pvc_size = None
    size_quantity = None
    # Try status.capacity (dict) - actual bound size
    capacity = getattr(getattr(pvc_obj, 'status', None), 'capacity', None)
    if isinstance(capacity, dict) and 'storage' in capacity:
        size_quantity = capacity['storage']
    # Fallback to spec.resources.requests (dict) - requested size
    if size_quantity is None:
        requests = getattr(getattr(pvc_obj.spec, 'resources', None), 'requests',
                           None)
        if isinstance(requests, dict):
            size_quantity = requests.get('storage')
    # Parse and normalize the size if found
    if size_quantity:
        try:
            # Normalize to GB string (e.g., '20')
            pvc_size = resources_utils.parse_memory_resource(
                size_quantity, 'size', allow_rounding=True)
        except ValueError as e:
            # Just log the error since it is not critical.
            logger.warning(f'Failed to parse PVC size {size_quantity!r} '
                           f'for PVC {pvc_name}: {e}')
    if pvc_size is not None:
        if config.size is not None and config.size != pvc_size:
            logger.warning(f'PVC {pvc_name} has size {pvc_size} but config '
                           f'size is {config.size}, overriding the config size'
                           f' with the PVC size.')
        config.size = pvc_size


def create_persistent_volume_claim(
    namespace: str,
    context: Optional[str],
    pvc_spec: Dict[str, Any],
    config: Optional[models.VolumeConfig] = None,
) -> None:
    """Creates a persistent volume claim for SkyServe controller."""
    pvc_name = pvc_spec['metadata']['name']
    try:
        pvc = kubernetes.core_api(
            context).read_namespaced_persistent_volume_claim(
                name=pvc_name,
                namespace=namespace,
                _request_timeout=kubernetes.API_TIMEOUT)
        if config is not None:
            _populate_config_from_pvc(config, pvc)
        logger.debug(f'PVC {pvc_name} already exists')
        return
    except kubernetes.api_exception() as e:
        if e.status != 404:  # Not found
            raise
    use_existing = config is not None and config.config.get('use_existing')
    if use_existing:
        raise ValueError(
            f'PVC {pvc_name} does not exist while use_existing is True.')
    pvc = kubernetes.core_api(
        context).create_namespaced_persistent_volume_claim(
            namespace=namespace,
            body=pvc_spec,
            _request_timeout=kubernetes.API_TIMEOUT)
    logger.info(f'Created PVC {pvc_name} in namespace {namespace}')
    if config is not None:
        _populate_config_from_pvc(config, pvc)


def _get_pvc_spec(namespace: str,
                  config: models.VolumeConfig) -> Dict[str, Any]:
    """Gets the PVC spec for the given storage config."""
    access_mode = config.config.get('access_mode')
    size = config.size
    # The previous code assumes that the access_mode and size are always set.
    assert access_mode is not None, f'access_mode is None for volume ' \
                                    f'{config.name_on_cloud}'
    pvc_spec: Dict[str, Any] = {
        'metadata': {
            'name': config.name_on_cloud,
            'namespace': namespace,
            'labels': {
                'parent': 'skypilot',
                'skypilot-name': config.name,
            }
        },
        'spec': {
            'accessModes': [access_mode],
        }
    }
    if size is not None:
        pvc_spec['spec']['resources'] = {'requests': {'storage': f'{size}Gi'}}
    if config.labels:
        pvc_spec['metadata']['labels'].update(config.labels)
    storage_class = config.config.get('storage_class_name')
    if storage_class is not None:
        pvc_spec['spec']['storageClassName'] = storage_class
    return pvc_spec



""" Digital Ocean Cloud. """

import json
import os
import typing
from typing import Dict, Iterator, List, Optional, Tuple, Union

from sky import catalog
from sky import clouds
from sky.adaptors import do
from sky.provision.do import utils as do_utils
from sky.utils import registry
from sky.utils import resources_utils

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib
    from sky.utils import volume as volume_lib

_CREDENTIAL_FILE = 'config.yaml'


@registry.CLOUD_REGISTRY.register(aliases=['digitalocean'])
class DO(clouds.Cloud):
    """Digital Ocean Cloud"""

    _REPR = 'DO'
    _CLOUD_UNSUPPORTED_FEATURES = {
        clouds.CloudImplementationFeatures.CLONE_DISK_FROM_CLUSTER:
            'Migrating '
            f'disk is not supported in {_REPR}.',
        clouds.CloudImplementationFeatures.SPOT_INSTANCE:
            'Spot instances are '
            f'not supported in {_REPR}.',
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER:
            'Custom disk tiers'
            f' is not supported in {_REPR}.',
        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER:
            ('Custom network tier is currently not supported in '
             f'{_REPR}.'),
        clouds.CloudImplementationFeatures.HIGH_AVAILABILITY_CONTROLLERS:
            ('High availability controllers are not supported in '
             f'{_REPR}.'),
        clouds.CloudImplementationFeatures.CUSTOM_MULTI_NETWORK:
            ('Customized multiple network interfaces are not supported in '
             f'{_REPR}.'),
        clouds.CloudImplementationFeatures.LOCAL_DISK:
            (f'Local disk is not supported on {_REPR}')
    }
    # DO maximum node name length defined as <= 255
    # https://docs.digitalocean.com/reference/api/api-reference/#operation/droplets_create
    # 255 - 8 = 247 characters since
    # our provisioner adds additional `-worker`.
    _MAX_CLUSTER_NAME_LEN_LIMIT = 247
    _regions: List[clouds.Region] = []

    # Using the latest SkyPilot provisioner API to provision and check status.
    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    @classmethod
    def _unsupported_features_for_resources(
        cls,
        resources: 'resources_lib.Resources',
        region: Optional[str] = None,
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        """The features not supported based on the resources provided.

        This method is used by check_features_are_supported() to check if the
        cloud implementation supports all the requested features.

        Returns:
            A dict of {feature: reason} for the features not supported by the
            cloud implementation.
        """
        del resources  # unused
        return cls._CLOUD_UNSUPPORTED_FEATURES

    @classmethod
    def _max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    def regions_with_offering(
        cls,
        instance_type: str,
        accelerators: Optional[Dict[str, int]],
        use_spot: bool,
        region: Optional[str],
        zone: Optional[str],
        resources: Optional['resources_lib.Resources'] = None,
    ) -> List[clouds.Region]:
        assert zone is None, 'DO does not support zones.'
        del accelerators, zone  # unused
        if use_spot:
            return []
        regions = catalog.get_region_zones_for_instance_type(
            instance_type, use_spot, 'DO')
        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        return catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                        clouds='DO')

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        del num_nodes  # unused
        regions = cls.regions_with_offering(instance_type,
                                            accelerators,
                                            use_spot,
                                            region=region,
                                            zone=None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    def instance_type_to_hourly_cost(
        self,
        instance_type: str,
        use_spot: bool,
        region: Optional[str] = None,
        zone: Optional[str] = None,
    ) -> float:
        return catalog.get_hourly_cost(
            instance_type,
            use_spot=use_spot,
            region=region,
            zone=zone,
            clouds='DO',
        )

    def accelerators_to_hourly_cost(
        self,
        accelerators: Dict[str, int],
        use_spot: bool,
        region: Optional[str] = None,
        zone: Optional[str] = None,
    ) -> float:
        """Returns the hourly cost of the accelerators, in dollars/hour."""
        # the acc price is include in the instance price.
        del accelerators, use_spot, region, zone  # unused
        return 0.0

    def get_egress_cost(self, num_gigabytes: float) -> float:
        return 0.0

    def __repr__(self):
        return self._REPR

    @classmethod
    def get_default_instance_type(cls,
                                  cpus: Optional[str] = None,
                                  memory: Optional[str] = None,
                                  disk_tier: Optional[
                                      resources_utils.DiskTier] = None,
                                  local_disk: Optional[str] = None,
                                  region: Optional[str] = None,
                                  zone: Optional[str] = None) -> Optional[str]:
        """Returns the default instance type for DO."""
        return catalog.get_default_instance_type(cpus=cpus,
                                                 memory=memory,
                                                 disk_tier=disk_tier,
                                                 local_disk=local_disk,
                                                 region=region,
                                                 zone=zone,
                                                 clouds='DO')

    @classmethod
    def get_accelerators_from_instance_type(
            cls, instance_type: str) -> Optional[Dict[str, Union[int, float]]]:
        return catalog.get_accelerators_from_instance_type(instance_type,
                                                           clouds='DO')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: resources_utils.ClusterName,
        region: 'clouds.Region',
        zones: Optional[List['clouds.Zone']],
        num_nodes: int,
        dryrun: bool = False,
        volume_mounts: Optional[List['volume_lib.VolumeMount']] = None,
    ) -> Dict[str, Optional[str]]:
        del zones, dryrun, cluster_name

        resources = resources.assert_launchable()
        acc_dict = self.get_accelerators_from_instance_type(
            resources.instance_type)
        if acc_dict is not None:
            custom_resources = json.dumps(acc_dict, separators=(',', ':'))
        else:
            custom_resources = None
        image_id = None
        if (resources.image_id is not None and
                resources.extract_docker_image() is None):
            if None in resources.image_id:
                image_id = resources.image_id[None]
            else:
                assert region.name in resources.image_id
                image_id = resources.image_id[region.name]
        return {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'region': region.name,
            **({
                'image_id': image_id
            } if image_id else {})
        }

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> resources_utils.FeasibleResources:
        """Returns a list of feasible resources for the given resources."""
        if resources.use_spot:
            # TODO: Add hints to all return values in this method to help
            #  users understand why the resources are not launchable.
            return resources_utils.FeasibleResources([], [], None)
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            resources = resources.copy(accelerators=None)
            return resources_utils.FeasibleResources([resources], [], None)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=DO(),
                    instance_type=instance_type,
                    accelerators=None,
                    cpus=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators
        if accelerators is None:
            # Return a default instance type
            default_instance_type = DO.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier,
                local_disk=resources.local_disk,
                region=resources.region,
                zone=resources.zone)
            if default_instance_type is None:
                return resources_utils.FeasibleResources([], [], None)
            else:
                return resources_utils.FeasibleResources(
                    _make([default_instance_type]), [], None)

        assert len(accelerators) == 1, resources
        acc, acc_count = list(accelerators.items())[0]
        (instance_list,
         fuzzy_candidate_list) = (catalog.get_instance_type_for_accelerator(
             acc,
             acc_count,
             use_spot=resources.use_spot,
             cpus=resources.cpus,
             memory=resources.memory,
             local_disk=resources.local_disk,
             region=resources.region,
             zone=resources.zone,
             clouds='DO',
         ))
        if instance_list is None:
            return resources_utils.FeasibleResources([], fuzzy_candidate_list,
                                                     None)
        return resources_utils.FeasibleResources(_make(instance_list),
                                                 fuzzy_candidate_list, None)

    @classmethod
    def _check_compute_credentials(
            cls) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Verify that the user has valid credentials for
        DO's compute service."""

        installed, err_msg = do.check_exceptions_dependencies_installed()
        if not installed:
            return False, err_msg

        try:
            # attempt to make a CURL request for listing instances
            do_utils.client().droplets.list()
        except do_utils.DigitalOceanError as err:
            return False, str(err)
        except do.exceptions().HttpResponseError as err:
            return False, str(err)

        return True, None

    def get_credential_file_mounts(self) -> Dict[str, str]:
        credential_path = do_utils.get_credentials_path()
        if credential_path is None:
            return {}
        if not os.path.exists(os.path.expanduser(credential_path)):
            return {}
        return {f'~/.config/doctl/{_CREDENTIAL_FILE}': credential_path}

    @classmethod
    def get_current_user_identity(cls) -> Optional[List[str]]:
        # NOTE: used for very advanced SkyPilot functionality
        # Can implement later if desired
        return None

    @classmethod
    def get_image_size(cls, image_id: str, region: Optional[str]) -> float:
        del region
        try:
            response = do_utils.client().images.get(image_id=image_id)
            return response['image']['size_gigabytes']
        except do.exceptions().HttpResponseError as err:
            raise do_utils.DigitalOceanError(
                'HTTP error while retrieving size of '
                f'image_id {response}: {err.error.message}') from err
        except KeyError as err:
            raise do_utils.DigitalOceanError(
                f'No image_id `{image_id}` found') from err

    def instance_type_exists(self, instance_type: str) -> bool:
        return catalog.instance_type_exists(instance_type, 'DO')

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        return catalog.validate_region_zone(region, zone, clouds='DO')



"""Vapi Connect
"""

import typing

from urllib3.exceptions import InsecureRequestWarning

from sky.adaptors import common as adaptors_common
from sky.adaptors import vsphere as vsphere_adaptor

if typing.TYPE_CHECKING:
    import requests
else:
    requests = adaptors_common.LazyImport('requests')


def get_jsonrpc_endpoint_url(host):
    # The URL for the stub requests are made against the /api HTTP endpoint
    # of the vCenter system.
    return 'https://{}/api'.format(host)


def connect(host,
            user,
            pwd,
            skip_verification=False,
            cert_path=None,
            suppress_warning=True):
    """Create an authenticated stub configuration object
    that can be used to issue
    requests against vCenter.

    Returns a stub_config that stores the session identifier that can be used
    to issue authenticated requests against vCenter.
    """
    host_url = get_jsonrpc_endpoint_url(host)

    session = requests.Session()
    if skip_verification:
        session = create_unverified_session(session, suppress_warning)
    elif cert_path:
        session.verify = cert_path
    connector = vsphere_adaptor.get_vapi_connect().get_requests_connector(
        session=session, url=host_url)
    stub_config = vsphere_adaptor.get_factories(
    ).StubConfigurationFactory.new_std_configuration(connector)

    return login(stub_config, user, pwd)


def login(stub_config, user, pwd):
    """Create an authenticated session with vCenter.

    Returns a stub_config that stores the session identifier that can be used
    to issue authenticated requests against vCenter.
    """
    # Pass user credentials (user/password) in the security context to
    # authenticate.
    user_password_security_context = vsphere_adaptor.get_user_password(
    ).create_user_password_security_context(user, pwd)
    stub_config.connector.set_security_context(user_password_security_context)

    # Create the stub for the session service and login by creating a session.
    session_svc = vsphere_adaptor.get_cis_client().Session(stub_config)
    session_id = session_svc.create()

    # Successful authentication.  Store the session identifier in the security
    # context of the stub and use that for all subsequent remote requests
    session_security_context = vsphere_adaptor.get_security_session(
    ).create_session_security_context(session_id)
    stub_config.connector.set_security_context(session_security_context)

    return stub_config


def logout(stub_config):
    """Delete session with vCenter.
    """
    if stub_config:
        session_svc = vsphere_adaptor.get_cis_client().Session(stub_config)
        session_svc.delete()


def create_unverified_session(session, suppress_warning=True):
    """Create a unverified session to disable the server certificate
    verification.
    This is not recommended in production code.
    """
    session.verify = False
    if suppress_warning:
        # Suppress unverified https request warnings
        requests.packages.urllib3.disable_warnings(  # type: ignore
            InsecureRequestWarning)
    return session



"""Cudo Compute"""
import subprocess
import typing
from typing import Dict, Iterator, List, Optional, Tuple, Union

from sky import catalog
from sky import clouds
from sky.adaptors import common
from sky.utils import common_utils
from sky.utils import registry
from sky.utils import resources_utils

if typing.TYPE_CHECKING:
    # Renaming to avoid shadowing variables.
    from sky import resources as resources_lib
    from sky.utils import volume as volume_lib

_CREDENTIAL_FILES = [
    # credential files for Cudo,
    'cudo.yml'
]


def _run_output(cmd):
    proc = subprocess.run(cmd,
                          shell=True,
                          check=True,
                          stderr=subprocess.PIPE,
                          stdout=subprocess.PIPE)
    return proc.stdout.decode('ascii')


@registry.CLOUD_REGISTRY.register
class Cudo(clouds.Cloud):
    """Cudo Compute"""
    _REPR = 'Cudo'

    _INDENT_PREFIX = '    '
    _DEPENDENCY_HINT = (
        'Cudo tools are not installed. Run the following commands:\n'
        f'{_INDENT_PREFIX}  $ pip install cudo-compute')

    _CREDENTIAL_HINT = (
        'Install cudoctl and run the following commands:\n'
        f'{_INDENT_PREFIX}  $ cudoctl init\n'
        f'{_INDENT_PREFIX}For more info: '
        # pylint: disable=line-too-long
        'https://docs.skypilot.co/en/latest/getting-started/installation.html')

    _PROJECT_HINT = (
        'Create a project and then set it as the default project,:\n'
        f'{_INDENT_PREFIX} $ cudoctl projects create my-project-name\n'
        f'{_INDENT_PREFIX} $ cudoctl init\n'
        f'{_INDENT_PREFIX}For more info: '
        # pylint: disable=line-too-long
        'https://docs.skypilot.co/en/latest/getting-started/installation.html')

    _CLOUD_UNSUPPORTED_FEATURES = {
        clouds.CloudImplementationFeatures.STOP: 'Stopping not supported.',
        clouds.CloudImplementationFeatures.SPOT_INSTANCE:
            ('Spot is not supported, as Cudo API does not implement spot.'),
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER:
            ('Custom disk tier is currently not supported on Cudo Compute'),
        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER:
            ('Custom network tier is currently not supported on Cudo Compute'),
        clouds.CloudImplementationFeatures.IMAGE_ID:
            ('Image ID is currently not supported on Cudo. '),
        clouds.CloudImplementationFeatures.DOCKER_IMAGE:
            ('Docker image is currently not supported on Cudo. You can try '
             'running docker command inside the `run` section in task.yaml.'),
        clouds.CloudImplementationFeatures.HOST_CONTROLLERS: (
            'Cudo Compute cannot host a controller as it does not '
            'autostopping, which will leave the controller to run indefinitely.'
        ),
        clouds.CloudImplementationFeatures.HIGH_AVAILABILITY_CONTROLLERS:
            ('High availability controllers are not supported on Cudo.'),
        clouds.CloudImplementationFeatures.CUSTOM_MULTI_NETWORK:
            ('Customized multiple network interfaces are not supported on Cudo.'
            ),
        clouds.CloudImplementationFeatures.LOCAL_DISK:
            (f'Local disk is not supported on {_REPR}')
    }
    _MAX_CLUSTER_NAME_LEN_LIMIT = 60

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    _regions: List[clouds.Region] = []

    @classmethod
    def _unsupported_features_for_resources(
        cls,
        resources: 'resources_lib.Resources',
        region: Optional[str] = None,
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        """The features not supported based on the resources provided.

        This method is used by check_features_are_supported() to check if the
        cloud implementation supports all the requested features.

        Returns:
            A dict of {feature: reason} for the features not supported by the
            cloud implementation.
        """
        del resources  # unused
        return cls._CLOUD_UNSUPPORTED_FEATURES

    @classmethod
    def _max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    def regions_with_offering(
        cls,
        instance_type,
        accelerators: Optional[Dict[str, int]],
        use_spot: bool,
        region: Optional[str],
        zone: Optional[str],
        resources: Optional['resources_lib.Resources'] = None,
    ) -> List[clouds.Region]:
        assert zone is None, 'Cudo does not support zones.'
        del accelerators, zone  # unused
        if use_spot:
            return []

        regions = catalog.get_region_zones_for_instance_type(
            instance_type, use_spot, 'cudo')

        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:

        return catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                        clouds='cudo')

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: Optional[str] = None,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        del num_nodes  # unused
        regions = cls.regions_with_offering(instance_type,
                                            accelerators,
                                            use_spot,
                                            region=region,
                                            zone=None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        return catalog.get_hourly_cost(instance_type,
                                       use_spot=use_spot,
                                       region=region,
                                       zone=zone,
                                       clouds='cudo')

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        del accelerators, use_spot, region, zone  # unused
        return 0.0

    def get_egress_cost(self, num_gigabytes: float) -> float:
        # Change if your cloud has egress cost. (You can do this later;
        # `return 0.0` is a good placeholder.)
        return 0.0

    @classmethod
    def get_default_instance_type(cls,
                                  cpus: Optional[str] = None,
                                  memory: Optional[str] = None,
                                  disk_tier: Optional[
                                      resources_utils.DiskTier] = None,
                                  local_disk: Optional[str] = None,
                                  region: Optional[str] = None,
                                  zone: Optional[str] = None) -> Optional[str]:
        return catalog.get_default_instance_type(cpus=cpus,
                                                 memory=memory,
                                                 disk_tier=disk_tier,
                                                 local_disk=local_disk,
                                                 region=region,
                                                 zone=zone,
                                                 clouds='cudo')

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, Union[int, float]]]:
        return catalog.get_accelerators_from_instance_type(instance_type,
                                                           clouds='cudo')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: resources_utils.ClusterName,
        region: 'clouds.Region',
        zones: Optional[List['clouds.Zone']],
        num_nodes: int,
        dryrun: bool = False,
        volume_mounts: Optional[List['volume_lib.VolumeMount']] = None,
    ) -> Dict[str, Optional[str]]:
        del zones, cluster_name  # unused
        resources = resources.assert_launchable()
        acc_dict = self.get_accelerators_from_instance_type(
            resources.instance_type)
        custom_resources = resources_utils.make_ray_custom_resources_str(
            acc_dict)

        return {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'region': region.name,
        }

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> 'resources_utils.FeasibleResources':
        if resources.use_spot:
            # TODO: Add hints to all return values in this method to help
            #  users understand why the resources are not launchable.
            return resources_utils.FeasibleResources([], [], None)
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            resources = resources.copy(accelerators=None)
            return resources_utils.FeasibleResources([resources], [], None)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=Cudo(),
                    instance_type=instance_type,
                    accelerators=None,
                    cpus=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators
        if accelerators is None:
            # Return a default instance type
            default_instance_type = Cudo.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier,
                local_disk=resources.local_disk,
                region=resources.region,
                zone=resources.zone)
            if default_instance_type is None:
                return resources_utils.FeasibleResources([], [], None)
            else:
                return resources_utils.FeasibleResources(
                    _make([default_instance_type]), [], None)

        assert len(accelerators) == 1, resources
        acc, acc_count = list(accelerators.items())[0]
        (instance_list,
         fuzzy_candidate_list) = catalog.get_instance_type_for_accelerator(
             acc,
             acc_count,
             use_spot=resources.use_spot,
             cpus=resources.cpus,
             memory=resources.memory,
             local_disk=resources.local_disk,
             region=resources.region,
             zone=resources.zone,
             clouds='cudo')
        if instance_list is None:
            return resources_utils.FeasibleResources([], fuzzy_candidate_list,
                                                     None)
        return resources_utils.FeasibleResources(_make(instance_list),
                                                 fuzzy_candidate_list, None)

    @classmethod
    def _check_compute_credentials(
            cls) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Checks if the user has access credentials to
        Cudo's compute service."""
        if not common.can_import_modules(['cudo_compute']):
            return False, (f'{cls._DEPENDENCY_HINT}\n'
                           f'{cls._INDENT_PREFIX}')

        try:
            _run_output('cudoctl --version')
        except (ImportError, subprocess.CalledProcessError) as e:
            return False, (
                f'{cls._CREDENTIAL_HINT}\n'
                f'{cls._INDENT_PREFIX}'
                f'{common_utils.format_exception(e, use_bracket=True)}')
        # pylint: disable=import-outside-toplevel,unused-import
        from cudo_compute import cudo_api
        from cudo_compute.rest import ApiException
        try:
            _, error = cudo_api.make_client()
        except FileNotFoundError as e:
            return False, (
                'Cudo credentials are not set. '
                f'{cls._CREDENTIAL_HINT}\n'
                f'{cls._INDENT_PREFIX}'
                f'{common_utils.format_exception(e, use_bracket=True)}')

        if error is not None:
            return False, (
                f'Application credentials are not set. '
                f'{common_utils.format_exception(error, use_bracket=True)}')

        project_id, error = cudo_api.get_project_id()
        if error is not None:
            return False, (
                f'Default project is not set. '
                f'{cls._PROJECT_HINT}\n'
                f'{cls._INDENT_PREFIX}'
                f'{common_utils.format_exception(error, use_bracket=True)}')
        try:
            api = cudo_api.virtual_machines()
            api.list_vms(project_id)
            return True, None
        except ApiException as e:
            return False, (
                f'Error calling API '
                f'{common_utils.format_exception(e, use_bracket=True)}')

    def get_credential_file_mounts(self) -> Dict[str, str]:
        return {
            f'~/.config/cudo/{filename}': f'~/.config/cudo/{filename}'
            for filename in _CREDENTIAL_FILES
        }

    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        # NOTE: used for very advanced SkyPilot functionality
        # Can implement later if desired
        return None

    def instance_type_exists(self, instance_type: str) -> bool:
        return catalog.instance_type_exists(instance_type, 'cudo')

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        return catalog.validate_region_zone(region, zone, clouds='cudo')



"""Constants for Kubernetes provisioning."""

# Canonical GPU names for GPU detection and labeling.
# Used by both GFDLabelFormatter and the GPU labeler script.
#
# IMPORTANT: Order matters for the GPU labeler script which uses substring
# matching (canonical_name in gpu_name). Names that are prefixes of other
# names must come later (e.g., 'L40S' before 'L40' before 'L4') to prevent
# 'L4' from matching 'L40S'. GFDLabelFormatter uses word boundary regex
# and is order-independent.
CANONICAL_GPU_NAMES = [
    # Blackwell architecture (2024+)
    'GB300',
    'GB200',
    'B300',
    'B200',
    'B100',
    # Hopper architecture
    'GH200',
    'H200',
    'H100-80GB',
    'H100-MEGA',
    'H100',
    # Ampere architecture
    'A100-80GB',
    'A100',
    'A10G',
    'A10',
    'A16',
    'A30',
    'A40',
    # Ada Lovelace architecture - Professional (RTX Ada)
    'RTX6000-Ada',
    'L40S',
    'L40',
    'L4',
    # Quadro/RTX Professional (Ampere)
    'A6000',
    'A5000',
    'A4000',
    # Older architectures - Volta/Pascal/Turing
    'V100',
    'P100',
    'P40',
    'P4000',
    'P4',
    'T4g',
    'T4',
    'K80',
    'M60',
]

NO_GPU_HELP_MESSAGE = ('If your cluster contains GPUs, make sure '
                       'nvidia.com/gpu resource is available on the nodes and '
                       'the node labels for identifying GPUs '
                       '(e.g., skypilot.co/accelerator) are setup correctly. ')

KUBERNETES_IN_CLUSTER_NAMESPACE_ENV_VAR = 'SKYPILOT_IN_CLUSTER_NAMESPACE'

# Name of kubernetes exec auth wrapper script
SKY_K8S_EXEC_AUTH_WRAPPER = 'sky-kube-exec-wrapper'

# PATH envvar for kubectl exec auth execve
SKY_K8S_EXEC_AUTH_PATH = '$HOME/skypilot-runtime/bin:$HOME/google-cloud-sdk/bin:$PATH'  # pylint: disable=line-too-long

# cache directory for kubeconfig with modified exec auth
SKY_K8S_EXEC_AUTH_KUBECONFIG_CACHE = '~/.sky/generated/kubeconfigs'

# Labels for the Pods created by SkyPilot
TAG_RAY_CLUSTER_NAME = 'ray-cluster-name'
TAG_POD_INITIALIZED = 'skypilot-initialized'
TAG_SKYPILOT_DEPLOYMENT_NAME = 'skypilot-deployment-name'

# Default name of the primary workload container in SkyPilot Ray pods.
RAY_NODE_CONTAINER_NAME = 'ray-node'

# Pod phases that are not holding PVCs
PVC_NOT_HOLD_POD_PHASES = ['Succeeded', 'Failed']



"""Add is_primary_in_job_group column to spot table.

This migration adds support for primary/auxiliary task markers in job groups:
- is_primary_in_job_group: Boolean indicating whether this task is "primary"
  (True) or "auxiliary" (False) within a job group. NULL for non-job-group
  jobs (single jobs and pipelines). When all primary tasks complete, auxiliary
  tasks are automatically terminated.

Revision ID: 014
Revises: 013
Create Date: 2026-01-19

"""
# pylint: disable=invalid-name
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa

from sky.utils.db import db_utils

# revision identifiers, used by Alembic.
revision: str = '014'
down_revision: Union[str, Sequence[str], None] = '013'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade():
    """Add is_primary_in_job_group column to spot table."""
    with op.get_context().autocommit_block():
        # Add is_primary_in_job_group column: Boolean indicating whether this
        # task is "primary" (True) or "auxiliary" (False) within a job group.
        # NULL for non-job-group jobs (single jobs and pipelines).
        db_utils.add_column_to_table_alembic('spot',
                                             'is_primary_in_job_group',
                                             sa.Boolean(),
                                             server_default=None)


def downgrade():
    """No downgrade logic."""
    pass



"""Async SDK for SkyServe."""
import asyncio
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

from sky.client import sdk_async
from sky.serve.client import sdk
from sky.usage import usage_lib

if typing.TYPE_CHECKING:
    import io

    import sky
    from sky.serve import serve_utils


@usage_lib.entrypoint
async def up(
    task: Union['sky.Task', 'sky.Dag'],
    service_name: str,
    # Internal only:
    # pylint: disable=invalid-name
    _need_confirmation: bool = False,
    stream_logs: Optional[
        sdk_async.StreamConfig] = sdk_async.DEFAULT_STREAM_CONFIG
) -> Tuple[str, str]:
    """Async version of up() that spins up a service."""
    request_id = await asyncio.to_thread(sdk.up, task, service_name,
                                         _need_confirmation)
    if stream_logs is not None:
        return await sdk_async._stream_and_get(request_id, stream_logs)  # pylint: disable=protected-access
    else:
        return await sdk_async.get(request_id)


@usage_lib.entrypoint
async def update(
    task: Union['sky.Task', 'sky.Dag'],
    service_name: str,
    mode: 'serve_utils.UpdateMode',
    # Internal only:
    # pylint: disable=invalid-name
    _need_confirmation: bool = False,
    stream_logs: Optional[
        sdk_async.StreamConfig] = sdk_async.DEFAULT_STREAM_CONFIG
) -> None:
    """Async version of update() that updates an existing service."""
    request_id = await asyncio.to_thread(sdk.update, task, service_name, mode,
                                         _need_confirmation)
    if stream_logs is not None:
        return await sdk_async._stream_and_get(request_id, stream_logs)  # pylint: disable=protected-access
    else:
        return await sdk_async.get(request_id)


@usage_lib.entrypoint
async def down(
    service_names: Optional[Union[str, List[str]]],
    all: bool = False,  # pylint: disable=redefined-builtin
    purge: bool = False,
    stream_logs: Optional[
        sdk_async.StreamConfig] = sdk_async.DEFAULT_STREAM_CONFIG
) -> None:
    """Async version of down() that tears down a service."""
    request_id = await asyncio.to_thread(sdk.down, service_names, all, purge)
    if stream_logs is not None:
        return await sdk_async._stream_and_get(request_id, stream_logs)  # pylint: disable=protected-access
    else:
        return await sdk_async.get(request_id)


@usage_lib.entrypoint
async def terminate_replica(
    service_name: str,
    replica_id: int,
    purge: bool,
    stream_logs: Optional[
        sdk_async.StreamConfig] = sdk_async.DEFAULT_STREAM_CONFIG
) -> None:
    """Async version of terminate_replica() that tears down a specific
    replica."""
    request_id = await asyncio.to_thread(sdk.terminate_replica, service_name,
                                         replica_id, purge)
    if stream_logs is not None:
        return await sdk_async._stream_and_get(request_id, stream_logs)  # pylint: disable=protected-access
    else:
        return await sdk_async.get(request_id)


@usage_lib.entrypoint
async def status(
    service_names: Optional[Union[str, List[str]]],
    stream_logs: Optional[
        sdk_async.StreamConfig] = sdk_async.DEFAULT_STREAM_CONFIG
) -> List[Dict[str, Any]]:
    """Async version of status() that sdk_async.gets service statuses."""
    request_id = await asyncio.to_thread(sdk.status, service_names)
    if stream_logs is not None:
        return await sdk_async._stream_and_get(request_id, stream_logs)  # pylint: disable=protected-access
    else:
        return await sdk_async.get(request_id)


@usage_lib.entrypoint
async def tail_logs(service_name: str,
                    target: Union[str, 'serve_utils.ServiceComponent'],
                    replica_id: Optional[int] = None,
                    follow: bool = True,
                    output_stream: Optional['io.TextIOBase'] = None) -> None:
    """Async version of tail_logs() that tails logs for a service."""
    return await asyncio.to_thread(sdk.tail_logs, service_name, target,
                                   replica_id, follow, output_stream)


@usage_lib.entrypoint
async def sync_down_logs(service_name: str,
                         local_dir: str,
                         *,
                         targets: Optional[Union[
                             str, 'serve_utils.ServiceComponent', List[Union[
                                 str, 'serve_utils.ServiceComponent']]]] = None,
                         replica_ids: Optional[List[int]] = None) -> None:
    """Async version of sync_down_logs() that syncs down logs from service
      components."""
    return await asyncio.to_thread(sdk.sync_down_logs,
                                   service_name,
                                   local_dir,
                                   targets=targets,
                                   replica_ids=replica_ids)



# Tests that require higher speed or could cause potential race conditions
# that cause pytest to fail when parallel enabled are put into this file.
# All tests in this file will not enable parallel.

import tempfile
import textwrap
import time

from click import testing as cli_testing

import sky
from sky import exceptions
from sky.client.cli import command
from sky.utils import registry


def _test_optimize_speed(resources: sky.Resources):
    with sky.Dag() as dag:
        task = sky.Task(run='echo hi')
        task.set_resources(resources)
    start = time.time()
    sky.optimize(dag)
    end = time.time()
    # 8.0 seconds = somewhat flaky.
    assert end - start < 8.0, (f'optimize took too long for {resources}, '
                               f'{end - start} seconds')


def test_optimize_speed(enable_all_clouds):
    _test_optimize_speed(sky.Resources(cpus=4))
    for cloud in registry.CLOUD_REGISTRY.values():
        _test_optimize_speed(sky.Resources(infra=str(cloud), cpus='4+'))
    _test_optimize_speed(sky.Resources(cpus='4+', memory='4+'))
    _test_optimize_speed(
        sky.Resources(cpus='4+', memory='4+', accelerators='V100:1'))
    _test_optimize_speed(
        sky.Resources(cpus='4+', memory='4+', accelerators='A100-80GB:8'))
    _test_optimize_speed(
        sky.Resources(cpus='4+', memory='4+', accelerators='tpu-v3-32'))


class TestAllCloudsEnabled:

    def test_accelerator_mismatch(self, enable_all_clouds):
        """Test the specified accelerator does not match the instance_type."""

        spec = textwrap.dedent("""\
            resources:
                cloud: aws
                instance_type: p3.2xlarge""")
        cli_runner = cli_testing.CliRunner()

        def _capture_mismatch_gpus_spec(file_path, gpus: str):
            result = cli_runner.invoke(command.launch,
                                       [file_path, '--gpus', gpus, '--dryrun'])
            assert isinstance(result.exception,
                              exceptions.ResourcesMismatchError)
            assert 'Infeasible resource demands found:' in str(result.exception)

        def _capture_match_gpus_spec(file_path, gpus: str):
            result = cli_runner.invoke(command.launch,
                                       [file_path, '--gpus', gpus, '--dryrun'])
            assert not result.exit_code

        with tempfile.NamedTemporaryFile('w', suffix='.yml') as f:
            f.write(spec)
            f.flush()

            _capture_mismatch_gpus_spec(f.name, 'T4:1')
            _capture_mismatch_gpus_spec(f.name, 'T4:0.5')
            _capture_mismatch_gpus_spec(f.name, 'V100:2')
            _capture_mismatch_gpus_spec(f.name, 'v100:2')
            _capture_mismatch_gpus_spec(f.name, 'V100:0.5')

            _capture_match_gpus_spec(f.name, 'V100:1')
            _capture_match_gpus_spec(f.name, 'V100:1')
            _capture_match_gpus_spec(f.name, 'V100')

    def test_k8s_alias(self, enable_all_clouds):
        cli_runner = cli_testing.CliRunner()

        result = cli_runner.invoke(command.launch,
                                   ['--cloud', 'k8s', '--dryrun'])
        assert not result.exit_code

        result = cli_runner.invoke(command.launch,
                                   ['--cloud', 'kubernetes', '--dryrun'])
        assert not result.exit_code

        result = cli_runner.invoke(command.launch,
                                   ['--cloud', 'notarealcloud', '--dryrun'])
        assert isinstance(result.exception, ValueError)

        result = cli_runner.invoke(command.gpus_list, ['--cloud', 'k8s'])
        assert not result.exit_code

        result = cli_runner.invoke(command.gpus_list, ['--cloud', 'kubernetes'])
        assert not result.exit_code

        result = cli_runner.invoke(command.gpus_list,
                                   ['--cloud', 'notarealcloud'])
        assert isinstance(result.exception, ValueError)



from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import ClassVar as _ClassVar, Iterable as _Iterable, Mapping as _Mapping, Optional as _Optional, Union as _Union

DESCRIPTOR: _descriptor.FileDescriptor

class JobStatus(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    JOB_STATUS_UNSPECIFIED: _ClassVar[JobStatus]
    JOB_STATUS_INIT: _ClassVar[JobStatus]
    JOB_STATUS_PENDING: _ClassVar[JobStatus]
    JOB_STATUS_SETTING_UP: _ClassVar[JobStatus]
    JOB_STATUS_RUNNING: _ClassVar[JobStatus]
    JOB_STATUS_FAILED_DRIVER: _ClassVar[JobStatus]
    JOB_STATUS_SUCCEEDED: _ClassVar[JobStatus]
    JOB_STATUS_FAILED: _ClassVar[JobStatus]
    JOB_STATUS_FAILED_SETUP: _ClassVar[JobStatus]
    JOB_STATUS_CANCELLED: _ClassVar[JobStatus]
JOB_STATUS_UNSPECIFIED: JobStatus
JOB_STATUS_INIT: JobStatus
JOB_STATUS_PENDING: JobStatus
JOB_STATUS_SETTING_UP: JobStatus
JOB_STATUS_RUNNING: JobStatus
JOB_STATUS_FAILED_DRIVER: JobStatus
JOB_STATUS_SUCCEEDED: JobStatus
JOB_STATUS_FAILED: JobStatus
JOB_STATUS_FAILED_SETUP: JobStatus
JOB_STATUS_CANCELLED: JobStatus

class AddJobRequest(_message.Message):
    __slots__ = ("job_name", "username", "run_timestamp", "resources_str", "metadata")
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    RUN_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_STR_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    job_name: str
    username: str
    run_timestamp: str
    resources_str: str
    metadata: str
    def __init__(self, job_name: _Optional[str] = ..., username: _Optional[str] = ..., run_timestamp: _Optional[str] = ..., resources_str: _Optional[str] = ..., metadata: _Optional[str] = ...) -> None: ...

class AddJobResponse(_message.Message):
    __slots__ = ("job_id", "log_dir")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    LOG_DIR_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    log_dir: str
    def __init__(self, job_id: _Optional[int] = ..., log_dir: _Optional[str] = ...) -> None: ...

class QueueJobRequest(_message.Message):
    __slots__ = ("job_id", "codegen", "script_path", "remote_log_dir", "managed_job")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    CODEGEN_FIELD_NUMBER: _ClassVar[int]
    SCRIPT_PATH_FIELD_NUMBER: _ClassVar[int]
    REMOTE_LOG_DIR_FIELD_NUMBER: _ClassVar[int]
    MANAGED_JOB_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    codegen: str
    script_path: str
    remote_log_dir: str
    managed_job: ManagedJobInfo
    def __init__(self, job_id: _Optional[int] = ..., codegen: _Optional[str] = ..., script_path: _Optional[str] = ..., remote_log_dir: _Optional[str] = ..., managed_job: _Optional[_Union[ManagedJobInfo, _Mapping]] = ...) -> None: ...

class ManagedJobInfo(_message.Message):
    __slots__ = ("name", "pool", "workspace", "entrypoint", "tasks", "user_id", "execution")
    NAME_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    ENTRYPOINT_FIELD_NUMBER: _ClassVar[int]
    TASKS_FIELD_NUMBER: _ClassVar[int]
    USER_ID_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_FIELD_NUMBER: _ClassVar[int]
    name: str
    pool: str
    workspace: str
    entrypoint: str
    tasks: _containers.RepeatedCompositeFieldContainer[ManagedJobTask]
    user_id: str
    execution: str
    def __init__(self, name: _Optional[str] = ..., pool: _Optional[str] = ..., workspace: _Optional[str] = ..., entrypoint: _Optional[str] = ..., tasks: _Optional[_Iterable[_Union[ManagedJobTask, _Mapping]]] = ..., user_id: _Optional[str] = ..., execution: _Optional[str] = ...) -> None: ...

class ManagedJobTask(_message.Message):
    __slots__ = ("task_id", "name", "resources_str", "metadata_json", "is_primary_in_job_group")
    TASK_ID_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_STR_FIELD_NUMBER: _ClassVar[int]
    METADATA_JSON_FIELD_NUMBER: _ClassVar[int]
    IS_PRIMARY_IN_JOB_GROUP_FIELD_NUMBER: _ClassVar[int]
    task_id: int
    name: str
    resources_str: str
    metadata_json: str
    is_primary_in_job_group: bool
    def __init__(self, task_id: _Optional[int] = ..., name: _Optional[str] = ..., resources_str: _Optional[str] = ..., metadata_json: _Optional[str] = ..., is_primary_in_job_group: bool = ...) -> None: ...

class QueueJobResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class UpdateStatusRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class UpdateStatusResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class GetJobQueueRequest(_message.Message):
    __slots__ = ("user_hash", "all_jobs")
    USER_HASH_FIELD_NUMBER: _ClassVar[int]
    ALL_JOBS_FIELD_NUMBER: _ClassVar[int]
    user_hash: str
    all_jobs: bool
    def __init__(self, user_hash: _Optional[str] = ..., all_jobs: bool = ...) -> None: ...

class JobInfo(_message.Message):
    __slots__ = ("job_id", "job_name", "username", "submitted_at", "status", "run_timestamp", "start_at", "end_at", "resources", "pid", "log_path", "metadata")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    JOB_NAME_FIELD_NUMBER: _ClassVar[int]
    USERNAME_FIELD_NUMBER: _ClassVar[int]
    SUBMITTED_AT_FIELD_NUMBER: _ClassVar[int]
    STATUS_FIELD_NUMBER: _ClassVar[int]
    RUN_TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    START_AT_FIELD_NUMBER: _ClassVar[int]
    END_AT_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_FIELD_NUMBER: _ClassVar[int]
    PID_FIELD_NUMBER: _ClassVar[int]
    LOG_PATH_FIELD_NUMBER: _ClassVar[int]
    METADATA_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    job_name: str
    username: str
    submitted_at: float
    status: JobStatus
    run_timestamp: str
    start_at: float
    end_at: float
    resources: str
    pid: int
    log_path: str
    metadata: str
    def __init__(self, job_id: _Optional[int] = ..., job_name: _Optional[str] = ..., username: _Optional[str] = ..., submitted_at: _Optional[float] = ..., status: _Optional[_Union[JobStatus, str]] = ..., run_timestamp: _Optional[str] = ..., start_at: _Optional[float] = ..., end_at: _Optional[float] = ..., resources: _Optional[str] = ..., pid: _Optional[int] = ..., log_path: _Optional[str] = ..., metadata: _Optional[str] = ...) -> None: ...

class GetJobQueueResponse(_message.Message):
    __slots__ = ("jobs",)
    JOBS_FIELD_NUMBER: _ClassVar[int]
    jobs: _containers.RepeatedCompositeFieldContainer[JobInfo]
    def __init__(self, jobs: _Optional[_Iterable[_Union[JobInfo, _Mapping]]] = ...) -> None: ...

class CancelJobsRequest(_message.Message):
    __slots__ = ("job_ids", "cancel_all", "user_hash")
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    CANCEL_ALL_FIELD_NUMBER: _ClassVar[int]
    USER_HASH_FIELD_NUMBER: _ClassVar[int]
    job_ids: _containers.RepeatedScalarFieldContainer[int]
    cancel_all: bool
    user_hash: str
    def __init__(self, job_ids: _Optional[_Iterable[int]] = ..., cancel_all: bool = ..., user_hash: _Optional[str] = ...) -> None: ...

class CancelJobsResponse(_message.Message):
    __slots__ = ("cancelled_job_ids",)
    CANCELLED_JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    cancelled_job_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, cancelled_job_ids: _Optional[_Iterable[int]] = ...) -> None: ...

class FailAllInProgressJobsRequest(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class FailAllInProgressJobsResponse(_message.Message):
    __slots__ = ()
    def __init__(self) -> None: ...

class TailLogsRequest(_message.Message):
    __slots__ = ("job_id", "managed_job_id", "follow", "tail")
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    MANAGED_JOB_ID_FIELD_NUMBER: _ClassVar[int]
    FOLLOW_FIELD_NUMBER: _ClassVar[int]
    TAIL_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    managed_job_id: int
    follow: bool
    tail: int
    def __init__(self, job_id: _Optional[int] = ..., managed_job_id: _Optional[int] = ..., follow: bool = ..., tail: _Optional[int] = ...) -> None: ...

class TailLogsResponse(_message.Message):
    __slots__ = ("log_line", "exit_code")
    LOG_LINE_FIELD_NUMBER: _ClassVar[int]
    EXIT_CODE_FIELD_NUMBER: _ClassVar[int]
    log_line: str
    exit_code: int
    def __init__(self, log_line: _Optional[str] = ..., exit_code: _Optional[int] = ...) -> None: ...

class GetJobStatusRequest(_message.Message):
    __slots__ = ("job_ids",)
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    job_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, job_ids: _Optional[_Iterable[int]] = ...) -> None: ...

class GetJobStatusResponse(_message.Message):
    __slots__ = ("job_statuses",)
    class JobStatusesEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: JobStatus
        def __init__(self, key: _Optional[int] = ..., value: _Optional[_Union[JobStatus, str]] = ...) -> None: ...
    JOB_STATUSES_FIELD_NUMBER: _ClassVar[int]
    job_statuses: _containers.ScalarMap[int, JobStatus]
    def __init__(self, job_statuses: _Optional[_Mapping[int, JobStatus]] = ...) -> None: ...

class GetJobSubmittedTimestampRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    def __init__(self, job_id: _Optional[int] = ...) -> None: ...

class GetJobSubmittedTimestampResponse(_message.Message):
    __slots__ = ("timestamp",)
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    timestamp: float
    def __init__(self, timestamp: _Optional[float] = ...) -> None: ...

class GetJobEndedTimestampRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    def __init__(self, job_id: _Optional[int] = ...) -> None: ...

class GetJobEndedTimestampResponse(_message.Message):
    __slots__ = ("timestamp",)
    TIMESTAMP_FIELD_NUMBER: _ClassVar[int]
    timestamp: float
    def __init__(self, timestamp: _Optional[float] = ...) -> None: ...

class GetLogDirsForJobsRequest(_message.Message):
    __slots__ = ("job_ids",)
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    job_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, job_ids: _Optional[_Iterable[int]] = ...) -> None: ...

class GetLogDirsForJobsResponse(_message.Message):
    __slots__ = ("job_log_dirs",)
    class JobLogDirsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: int
        value: str
        def __init__(self, key: _Optional[int] = ..., value: _Optional[str] = ...) -> None: ...
    JOB_LOG_DIRS_FIELD_NUMBER: _ClassVar[int]
    job_log_dirs: _containers.ScalarMap[int, str]
    def __init__(self, job_log_dirs: _Optional[_Mapping[int, str]] = ...) -> None: ...

class GetJobExitCodesRequest(_message.Message):
    __slots__ = ("job_id",)
    JOB_ID_FIELD_NUMBER: _ClassVar[int]
    job_id: int
    def __init__(self, job_id: _Optional[int] = ...) -> None: ...

class GetJobExitCodesResponse(_message.Message):
    __slots__ = ("exit_codes",)
    EXIT_CODES_FIELD_NUMBER: _ClassVar[int]
    exit_codes: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, exit_codes: _Optional[_Iterable[int]] = ...) -> None: ...

class SetJobInfoWithoutJobIdRequest(_message.Message):
    __slots__ = ("name", "workspace", "entrypoint", "pool", "pool_hash", "user_hash", "task_ids", "task_names", "resources_str", "metadata_jsons", "num_jobs", "execution", "is_primary_in_job_groups")
    NAME_FIELD_NUMBER: _ClassVar[int]
    WORKSPACE_FIELD_NUMBER: _ClassVar[int]
    ENTRYPOINT_FIELD_NUMBER: _ClassVar[int]
    POOL_FIELD_NUMBER: _ClassVar[int]
    POOL_HASH_FIELD_NUMBER: _ClassVar[int]
    USER_HASH_FIELD_NUMBER: _ClassVar[int]
    TASK_IDS_FIELD_NUMBER: _ClassVar[int]
    TASK_NAMES_FIELD_NUMBER: _ClassVar[int]
    RESOURCES_STR_FIELD_NUMBER: _ClassVar[int]
    METADATA_JSONS_FIELD_NUMBER: _ClassVar[int]
    NUM_JOBS_FIELD_NUMBER: _ClassVar[int]
    EXECUTION_FIELD_NUMBER: _ClassVar[int]
    IS_PRIMARY_IN_JOB_GROUPS_FIELD_NUMBER: _ClassVar[int]
    name: str
    workspace: str
    entrypoint: str
    pool: str
    pool_hash: str
    user_hash: str
    task_ids: _containers.RepeatedScalarFieldContainer[int]
    task_names: _containers.RepeatedScalarFieldContainer[str]
    resources_str: str
    metadata_jsons: _containers.RepeatedScalarFieldContainer[str]
    num_jobs: int
    execution: str
    is_primary_in_job_groups: _containers.RepeatedScalarFieldContainer[bool]
    def __init__(self, name: _Optional[str] = ..., workspace: _Optional[str] = ..., entrypoint: _Optional[str] = ..., pool: _Optional[str] = ..., pool_hash: _Optional[str] = ..., user_hash: _Optional[str] = ..., task_ids: _Optional[_Iterable[int]] = ..., task_names: _Optional[_Iterable[str]] = ..., resources_str: _Optional[str] = ..., metadata_jsons: _Optional[_Iterable[str]] = ..., num_jobs: _Optional[int] = ..., execution: _Optional[str] = ..., is_primary_in_job_groups: _Optional[_Iterable[bool]] = ...) -> None: ...

class SetJobInfoWithoutJobIdResponse(_message.Message):
    __slots__ = ("job_ids",)
    JOB_IDS_FIELD_NUMBER: _ClassVar[int]
    job_ids: _containers.RepeatedScalarFieldContainer[int]
    def __init__(self, job_ids: _Optional[_Iterable[int]] = ...) -> None: ...



"""Hyperbolic Cloud service catalog.

This module loads and queries the service catalog for Hyperbolic Cloud.
"""
from typing import Dict, List, Optional, Tuple, Union

from sky.catalog import common
from sky.clouds import cloud  # Import cloud here for Region
from sky.utils import ux_utils

# Initialize cloud variable at module level
CLOUD = 'hyperbolic'

_df = common.read_catalog('hyperbolic/vms.csv')


def instance_type_exists(instance_type: str) -> bool:
    return common.instance_type_exists_impl(_df, instance_type)


def validate_region_zone(
        region: Optional[str],
        zone: Optional[str]) -> Tuple[Optional[str], Optional[str]]:
    if zone is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Hyperbolic Cloud does not support zones.')
    return common.validate_region_zone_impl('hyperbolic', _df, region, zone)


def get_hourly_cost(
    instance_type: str,
    use_spot: bool = False,
    region: Optional[str] = None,
    zone: Optional[str] = None,
) -> float:
    if zone is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Hyperbolic Cloud does not support zones.')
    return common.get_hourly_cost_impl(_df, instance_type, use_spot, region,
                                       zone)


def get_vcpus_mem_from_instance_type(
    instance_type: str,) -> Tuple[Optional[float], Optional[float]]:
    return common.get_vcpus_mem_from_instance_type_impl(_df, instance_type)


def get_accelerators_from_instance_type(
        instance_type: str) -> Optional[Dict[str, Union[int, float]]]:
    return common.get_accelerators_from_instance_type_impl(_df, instance_type)


def get_vcpus_from_instance_type(instance_type: str) -> Optional[float]:
    vcpus, _ = get_vcpus_mem_from_instance_type(instance_type)
    return vcpus


def get_memory_from_instance_type(instance_type: str) -> Optional[float]:
    _, mem = get_vcpus_mem_from_instance_type(instance_type)
    return mem


def get_zone_shell_cmd() -> Optional[str]:
    """Returns the shell command to obtain the zone."""
    return None


def get_default_instance_type(cpus: Optional[str] = None,
                              memory: Optional[str] = None,
                              disk_tier: Optional[str] = None,
                              local_disk: Optional[str] = None,
                              region: Optional[str] = None,
                              zone: Optional[str] = None) -> Optional[str]:
    del disk_tier, local_disk  # Unused
    return common.get_instance_type_for_cpus_mem_impl(_df, cpus, memory, region,
                                                      zone)


def get_instance_type_for_accelerator(
    acc_name: str,
    acc_count: int,
    cpus: Optional[str] = None,
    memory: Optional[str] = None,
    use_spot: bool = False,
    local_disk: Optional[str] = None,
    region: Optional[str] = None,
    zone: Optional[str] = None,
) -> Tuple[Optional[List[str]], List[str]]:
    del local_disk  # unused
    if zone is not None:
        with ux_utils.print_exception_no_traceback():
            raise ValueError('Hyperbolic Cloud does not support zones.')
    return common.get_instance_type_for_accelerator_impl(df=_df,
                                                         acc_name=acc_name,
                                                         acc_count=acc_count,
                                                         cpus=cpus,
                                                         memory=memory,
                                                         use_spot=use_spot,
                                                         region=region,
                                                         zone=zone)


def get_region_zones_for_instance_type(instance_type: str,
                                       use_spot: bool) -> List[cloud.Region]:
    df = _df[_df['InstanceType'] == instance_type]
    return common.get_region_zones(df, use_spot)


def get_gen_version(instance_type: str) -> Optional[str]:
    """Returns the generation version of the instance type."""
    del instance_type  # Unused
    # TODO: Implement generation version detection
    return None


def list_accelerators(
    gpus_only: bool = True,
    name_filter: Optional[str] = None,
    region_filter: Optional[str] = None,
    quantity_filter: Optional[int] = None,
    case_sensitive: bool = True,
    all_regions: bool = False,
    require_price: bool = True,
) -> Dict[str, List[common.InstanceTypeInfo]]:
    """Returns all instance types in Hyperbolic Cloud offering accelerators."""
    del require_price  # Unused
    return common.list_accelerators_impl('Hyperbolic', _df, gpus_only,
                                         name_filter, region_filter,
                                         quantity_filter, case_sensitive,
                                         all_regions)


def get_instance_type_from_catalog() -> dict:
    # TODO: Implement this function
    return {}


def regions() -> List[cloud.Region]:
    return [cloud.Region('default')]



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc
import warnings

from sky.schemas.generated import managed_jobsv1_pb2 as sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2

GRPC_GENERATED_VERSION = '1.63.0'
GRPC_VERSION = grpc.__version__
EXPECTED_ERROR_RELEASE = '1.65.0'
SCHEDULED_RELEASE_DATE = 'June 25, 2024'
_version_not_supported = False

try:
    from grpc._utilities import first_version_is_lower
    _version_not_supported = first_version_is_lower(GRPC_VERSION, GRPC_GENERATED_VERSION)
except ImportError:
    _version_not_supported = True

if _version_not_supported:
    warnings.warn(
        f'The grpc package installed is at version {GRPC_VERSION},'
        + f' but the generated code in sky/schemas/generated/managed_jobsv1_pb2_grpc.py depends on'
        + f' grpcio>={GRPC_GENERATED_VERSION}.'
        + f' Please upgrade your grpc module to grpcio>={GRPC_GENERATED_VERSION}'
        + f' or downgrade your generated code using grpcio-tools<={GRPC_VERSION}.'
        + f' This warning will become an error in {EXPECTED_ERROR_RELEASE},'
        + f' scheduled for release on {SCHEDULED_RELEASE_DATE}.',
        RuntimeWarning
    )


class ManagedJobsServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetVersion = channel.unary_unary(
                '/managed_jobs.v1.ManagedJobsService/GetVersion',
                request_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetVersionRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetVersionResponse.FromString,
                _registered_method=True)
        self.GetJobTable = channel.unary_unary(
                '/managed_jobs.v1.ManagedJobsService/GetJobTable',
                request_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetJobTableRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetJobTableResponse.FromString,
                _registered_method=True)
        self.GetAllJobIdsByName = channel.unary_unary(
                '/managed_jobs.v1.ManagedJobsService/GetAllJobIdsByName',
                request_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetAllJobIdsByNameRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetAllJobIdsByNameResponse.FromString,
                _registered_method=True)
        self.CancelJobs = channel.unary_unary(
                '/managed_jobs.v1.ManagedJobsService/CancelJobs',
                request_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.CancelJobsRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.CancelJobsResponse.FromString,
                _registered_method=True)
        self.StreamLogs = channel.unary_stream(
                '/managed_jobs.v1.ManagedJobsService/StreamLogs',
                request_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.StreamLogsRequest.SerializeToString,
                response_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.StreamLogsResponse.FromString,
                _registered_method=True)


class ManagedJobsServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetVersion(self, request, context):
        """Get controller version.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetJobTable(self, request, context):
        """Get the managed job queue with advanced filtering.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetAllJobIdsByName(self, request, context):
        """Get all job IDs by name.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CancelJobs(self, request, context):
        """Cancel managed jobs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StreamLogs(self, request, context):
        """Stream managed job logs.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_ManagedJobsServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetVersion': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVersion,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetVersionRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetVersionResponse.SerializeToString,
            ),
            'GetJobTable': grpc.unary_unary_rpc_method_handler(
                    servicer.GetJobTable,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetJobTableRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetJobTableResponse.SerializeToString,
            ),
            'GetAllJobIdsByName': grpc.unary_unary_rpc_method_handler(
                    servicer.GetAllJobIdsByName,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetAllJobIdsByNameRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetAllJobIdsByNameResponse.SerializeToString,
            ),
            'CancelJobs': grpc.unary_unary_rpc_method_handler(
                    servicer.CancelJobs,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.CancelJobsRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.CancelJobsResponse.SerializeToString,
            ),
            'StreamLogs': grpc.unary_stream_rpc_method_handler(
                    servicer.StreamLogs,
                    request_deserializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.StreamLogsRequest.FromString,
                    response_serializer=sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.StreamLogsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'managed_jobs.v1.ManagedJobsService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class ManagedJobsService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetVersion(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/managed_jobs.v1.ManagedJobsService/GetVersion',
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetVersionRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetVersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetJobTable(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/managed_jobs.v1.ManagedJobsService/GetJobTable',
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetJobTableRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetJobTableResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetAllJobIdsByName(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/managed_jobs.v1.ManagedJobsService/GetAllJobIdsByName',
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetAllJobIdsByNameRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.GetAllJobIdsByNameResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CancelJobs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/managed_jobs.v1.ManagedJobsService/CancelJobs',
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.CancelJobsRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.CancelJobsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def StreamLogs(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/managed_jobs.v1.ManagedJobsService/StreamLogs',
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.StreamLogsRequest.SerializeToString,
            sky_dot_schemas_dot_generated_dot_managed__jobsv1__pb2.StreamLogsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



"""Utilities for nested config."""
import copy
from typing import Any, Dict, List, Optional, Tuple

from sky import sky_logging

logger = sky_logging.init_logger(__name__)

_REGION_CONFIG_CLOUDS = ['nebius', 'oci']

# Kubernetes API use list to represent dictionary fields with patch strategy
# merge and each item is indexed by the patch merge key. The following map
# maps the field name to the patch merge key.
# pylint: disable=line-too-long
# Ref: https://kubernetes.io/docs/reference/generated/kubernetes-api/v1.27/#podspec-v1-core
# NOTE: field imagePullSecrets are not included deliberately for backward compatibility
_PATCH_MERGE_KEYS = {
    'containers': 'name',
    'initContainers': 'name',
    'ephemeralContainers': 'name',
    'volumes': 'name',
    'volumeMounts': 'name',
    'resourceClaims': 'name',
    'env': 'name',
    'hostAliases': 'ip',
    'topologySpreadConstraints': 'topologyKey',
    'ports': 'containerPort',
    'volumeDevices': 'devicePath',
}


class Config(Dict[str, Any]):
    """SkyPilot config that supports setting/getting values with nested keys."""

    def get_nested(
        self,
        keys: Tuple[str, ...],
        default_value: Any,
        override_configs: Optional[Dict[str, Any]] = None,
        allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
        disallowed_override_keys: Optional[List[Tuple[str,
                                                      ...]]] = None) -> Any:
        """Gets a nested key.

        If any key is not found, or any intermediate key does not point to a
        dict value, returns 'default_value'.

        Args:
            keys: A tuple of strings representing the nested keys.
            default_value: The default value to return if the key is not found.
            override_configs: A dict of override configs with the same schema as
                the config file, but only containing the keys to override.
            allowed_override_keys: A list of keys that are allowed to be
                overridden.
            disallowed_override_keys: A list of keys that are disallowed to be
                overridden.

        Returns:
            The value of the nested key, or 'default_value' if not found.
        """
        config = copy.deepcopy(self)
        if override_configs is not None:
            config = _recursive_update(config, override_configs,
                                       allowed_override_keys,
                                       disallowed_override_keys)
        return _get_nested(config, keys, default_value, pop=False)

    def set_nested(self, keys: Tuple[str, ...], value: Any) -> None:
        """In-place sets a nested key to value.

        Like get_nested(), if any key is not found, this will not raise an
        error.
        """
        override = {}
        for i, key in enumerate(reversed(keys)):
            if i == 0:
                override = {key: value}
            else:
                override = {key: override}
        _recursive_update(self, override)

    def pop_nested(self, keys: Tuple[str, ...], default_value: Any) -> Any:
        """Pops a nested key."""
        return _get_nested(self, keys, default_value, pop=True)

    @classmethod
    def from_dict(cls, config: Optional[Dict[str, Any]]) -> 'Config':
        if config is None:
            return cls()
        return cls(**config)


def _check_allowed_and_disallowed_override_keys(
    key: str,
    allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
    disallowed_override_keys: Optional[List[Tuple[str, ...]]] = None
) -> Tuple[Optional[List[Tuple[str, ...]]], Optional[List[Tuple[str, ...]]]]:
    allowed_keys_with_matched_prefix: Optional[List[Tuple[str, ...]]] = []
    disallowed_keys_with_matched_prefix: Optional[List[Tuple[str, ...]]] = []
    if allowed_override_keys is not None:
        for nested_key in allowed_override_keys:
            if key == nested_key[0]:
                if len(nested_key) == 1:
                    # Allowed key is fully matched, no need to check further.
                    allowed_keys_with_matched_prefix = None
                    break
                assert allowed_keys_with_matched_prefix is not None
                allowed_keys_with_matched_prefix.append(nested_key[1:])
        if (allowed_keys_with_matched_prefix is not None and
                not allowed_keys_with_matched_prefix):
            raise ValueError(f'Key {key} is not in allowed override keys: '
                             f'{allowed_override_keys}')
    else:
        allowed_keys_with_matched_prefix = None

    if disallowed_override_keys is not None:
        for nested_key in disallowed_override_keys:
            if key == nested_key[0]:
                if len(nested_key) == 1:
                    raise ValueError(
                        f'Key {key} is in disallowed override keys: '
                        f'{disallowed_override_keys}')
                assert disallowed_keys_with_matched_prefix is not None
                disallowed_keys_with_matched_prefix.append(nested_key[1:])
    else:
        disallowed_keys_with_matched_prefix = None
    return allowed_keys_with_matched_prefix, disallowed_keys_with_matched_prefix


def _recursive_update(
        base_config: Config,
        override_config: Dict[str, Any],
        allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
        disallowed_override_keys: Optional[List[Tuple[str,
                                                      ...]]] = None) -> Config:
    """Recursively updates base configuration with override configuration"""

    def _update_k8s_config(
        base_config: Config,
        override_config: Dict[str, Any],
        allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
        disallowed_override_keys: Optional[List[Tuple[str,
                                                      ...]]] = None) -> Config:
        """Updates the top-level k8s config with the override config."""
        for key, value in override_config.items():
            (next_allowed_override_keys, next_disallowed_override_keys
            ) = _check_allowed_and_disallowed_override_keys(
                key, allowed_override_keys, disallowed_override_keys)
            if key in ['custom_metadata', 'pod_config'] and key in base_config:
                merge_k8s_configs(base_config[key], value,
                                  next_allowed_override_keys,
                                  next_disallowed_override_keys)
            elif (isinstance(value, dict) and key in base_config and
                  isinstance(base_config[key], dict)):
                _recursive_update(base_config[key], value,
                                  next_allowed_override_keys,
                                  next_disallowed_override_keys)
            else:
                base_config[key] = value
        return base_config

    for key, value in override_config.items():
        (next_allowed_override_keys, next_disallowed_override_keys
        ) = _check_allowed_and_disallowed_override_keys(
            key, allowed_override_keys, disallowed_override_keys)
        if key == 'kubernetes' and key in base_config:
            _update_k8s_config(base_config[key], value,
                               next_allowed_override_keys,
                               next_disallowed_override_keys)
        elif (isinstance(value, dict) and key in base_config and
              isinstance(base_config[key], dict)):
            _recursive_update(base_config[key], value,
                              next_allowed_override_keys,
                              next_disallowed_override_keys)
        else:
            base_config[key] = value
    return base_config


def _get_nested(configs: Optional[Dict[str, Any]],
                keys: Tuple[str, ...],
                default_value: Any,
                pop: bool = False) -> Any:
    if configs is None:
        return default_value
    curr = configs
    for i, key in enumerate(keys):
        if isinstance(curr, dict) and key in curr:
            value = curr[key]
            if i == len(keys) - 1:
                if pop:
                    curr.pop(key, default_value)
            curr = value
        else:
            return default_value
    return curr


def merge_k8s_configs(
        base_config: Dict[Any, Any],
        override_config: Dict[Any, Any],
        allowed_override_keys: Optional[List[Tuple[str, ...]]] = None,
        disallowed_override_keys: Optional[List[Tuple[str,
                                                      ...]]] = None) -> None:
    """Merge two configs into the base_config.

    Updates nested dictionaries instead of replacing them.
    If a list is encountered, it will be appended to the base_config list.

    For fields with Kubernetes patch merge strategy (containers, volumes, env,
    etc.), items are merged by their patch merge key (e.g., 'name' for
    containers). If an item with the same key exists in base_config, it is
    merged; otherwise, the new item is appended.

    Special handling for 'containers': for backward compatibility, if a
    container in the override does not have a 'name' field, it is merged
    into the first container in the base config (legacy behavior). If the
    container has a 'name' field, patch merge by name is used.
    """
    for key, value in override_config.items():
        (next_allowed_override_keys, next_disallowed_override_keys
        ) = _check_allowed_and_disallowed_override_keys(
            key, allowed_override_keys, disallowed_override_keys)
        if isinstance(value, dict) and key in base_config:
            merge_k8s_configs(base_config[key], value,
                              next_allowed_override_keys,
                              next_disallowed_override_keys)
        elif isinstance(value, list) and key in base_config:
            assert isinstance(base_config[key], list), \
                f'Expected {key} to be a list, found {base_config[key]}'
            if key == 'imagePullSecrets':
                # For imagePullSecrets, merge the first item from override
                # into the first item in base (legacy behavior).
                assert len(value) == 1, \
                    f'Expected only one imagePullSecret, found {value}'
                merge_k8s_configs(base_config[key][0], value[0],
                                  next_allowed_override_keys,
                                  next_disallowed_override_keys)
            # For list fields with patch strategy "merge", we merge the list
            # by the patch merge key.
            elif key in _PATCH_MERGE_KEYS:
                patch_merge_key = _PATCH_MERGE_KEYS[key]
                for override_item in value:
                    override_item_name = override_item.get(patch_merge_key)
                    if override_item_name is not None:
                        # Item has a name - use patch merge by name
                        existing_base_item = next(
                            (v for v in base_config[key]
                             if v.get(patch_merge_key) == override_item_name),
                            None)
                        if existing_base_item is not None:
                            merge_k8s_configs(existing_base_item, override_item)
                        else:
                            base_config[key].append(override_item)
                    elif key == 'containers' and base_config[key]:
                        # Backward compatibility for containers: if no name is
                        # specified, merge into the first container (index 0)
                        merge_k8s_configs(base_config[key][0], override_item,
                                          next_allowed_override_keys,
                                          next_disallowed_override_keys)
                    else:
                        base_config[key].append(override_item)
            else:
                base_config[key].extend(value)
        else:
            base_config[key] = value


def get_cloud_config_value_from_dict(
        dict_config: Dict[str, Any],
        cloud: str,
        keys: Tuple[str, ...],
        region: Optional[str] = None,
        default_value: Optional[Any] = None,
        override_configs: Optional[Dict[str, Any]] = None) -> Any:
    """Returns the nested key value by reading from config
    Order to get the property_name value:
    1. if region is specified,
       try to get the value from <cloud>/<region_key>/<region>/keys
    2. if no region or no override,
       try to get it at the cloud level <cloud>/keys
    3. if not found at cloud level,
       return either default_value if specified or None
    """
    input_config = Config(dict_config)
    region_key = None
    if cloud in ('kubernetes', 'ssh'):
        region_key = 'context_configs'
    elif cloud in _REGION_CONFIG_CLOUDS:
        region_key = 'region_configs'

    per_context_config = None
    if region is not None and region_key is not None:
        per_context_config = input_config.get_nested(
            keys=(cloud, region_key, region) + keys,
            default_value=None,
            override_configs=override_configs)
    # if no override found for specified region
    general_config = input_config.get_nested(keys=(cloud,) + keys,
                                             default_value=default_value,
                                             override_configs=override_configs)

    if (cloud == 'kubernetes' and isinstance(general_config, dict) and
            isinstance(per_context_config, dict)):
        merge_k8s_configs(general_config, per_context_config)
        return general_config
    else:
        return (general_config
                if per_context_config is None else per_context_config)



"""Data Models for SkyPilot."""

import collections
import dataclasses
import enum
import getpass
import os
from typing import Any, ClassVar, Dict, List, Optional

import pydantic

from sky.skylet import constants
from sky.utils import common_utils


class UserType(enum.Enum):
    """Enum for user types."""
    # Internal system users (SERVER_ID, SKYPILOT_SYSTEM_USER_ID)
    SYSTEM = 'system'
    # Users authenticated by basic auth on the API server that have a password
    BASIC = 'basic'
    # Service accounts
    SA = 'sa'
    # Users authenticated via SSO
    SSO = 'sso'
    # Users authenticated by basic auth on the ingress that have no password
    LEGACY = 'legacy'


@dataclasses.dataclass
class User:
    """Dataclass to store user information."""
    # User hash
    id: str
    # Display name of the user
    name: Optional[str] = None
    password: Optional[str] = None
    created_at: Optional[int] = None
    user_type: Optional[str] = None

    def __init__(
        self,
        id: str,  # pylint: disable=redefined-builtin
        name: Optional[str] = None,
        password: Optional[str] = None,
        created_at: Optional[int] = None,
        user_type: Optional[str] = None,
    ):
        self.id = id.strip().lower()
        self.name = name
        self.password = password
        self.created_at = created_at
        self.user_type = user_type

    def to_dict(self) -> Dict[str, Any]:
        return {'id': self.id, 'name': self.name, 'user_type': self.user_type}

    @classmethod
    def get_current_user(cls) -> 'User':
        """Returns the current user."""
        user_name = os.getenv(constants.USER_ENV_VAR, getpass.getuser())
        user_hash = common_utils.get_user_hash()
        return User(id=user_hash, name=user_name)

    def is_service_account(self) -> bool:
        """Check if the user is a service account."""
        return self.id.lower().startswith('sa-')


RealtimeGpuAvailability = collections.namedtuple(
    'RealtimeGpuAvailability', ['gpu', 'counts', 'capacity', 'available'])


@dataclasses.dataclass
class KubernetesNodeInfo:
    """Dataclass to store Kubernetes node information."""
    name: str
    accelerator_type: Optional[str]
    # Resources available on the node. E.g., {'nvidia.com/gpu': '2'}
    total: Dict[str, int]
    free: Dict[str, int]
    # IP address of the node (external IP preferred, fallback to internal IP)
    ip_address: Optional[str] = None
    # CPU count (total CPUs available on the node)
    cpu_count: Optional[float] = None
    # Memory in GB (total memory available on the node)
    memory_gb: Optional[float] = None
    # Free CPU count (free CPUs available on the node after pod allocations)
    cpu_free: Optional[float] = None
    # Free memory in GB (free memory available on the node after pod
    # allocations)
    memory_free_gb: Optional[float] = None
    # Whether the node is ready (all conditions are satisfied)
    is_ready: bool = True
    # Whether the node is cordoned (spec.unschedulable is true)
    is_cordoned: bool = False
    # List of taints on the node, each taint is a dict with 'key', 'value',
    # 'effect'
    taints: Optional[List[Dict[str, Any]]] = None


@dataclasses.dataclass
class KubernetesNodesInfo:
    """Dataclass to store Kubernetes node info map."""
    # The nodes in the cluster, keyed by node name.
    node_info_dict: Dict[str, KubernetesNodeInfo]
    # Additional hint for the node info.
    hint: str

    def to_dict(self) -> Dict[str, Any]:
        return {
            'node_info_dict': {
                node_name: dataclasses.asdict(node_info)
                for node_name, node_info in self.node_info_dict.items()
            },
            'hint': self.hint,
        }

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'KubernetesNodesInfo':
        return cls(
            node_info_dict={
                node_name: KubernetesNodeInfo(**node_info)
                for node_name, node_info in data['node_info_dict'].items()
            },
            hint=data['hint'],
        )


class VolumeConfig(pydantic.BaseModel):
    """Configuration for creating a volume."""
    # If any fields changed, increment the version. For backward compatibility,
    # modify the __setstate__ method to handle the old version.
    _VERSION: ClassVar[int] = 1

    _version: int
    name: str
    type: str
    cloud: str
    region: Optional[str]
    zone: Optional[str]
    name_on_cloud: str
    size: Optional[str]
    config: Dict[str, Any] = {}
    labels: Optional[Dict[str, str]] = None
    id_on_cloud: Optional[str] = None

    def __getstate__(self) -> Dict[str, Any]:
        state = super().__getstate__()
        state['_version'] = self._VERSION
        return state

    def __setstate__(self, state: Dict[str, Any]) -> None:
        """Set state from pickled state, for backward compatibility."""
        super().__setstate__(state)
        version = state.pop('_version', None)
        if version is None:
            version = -1

        if version < 0:
            state['id_on_cloud'] = None

        state['_version'] = self._VERSION
        self.__dict__.update(state)
